'''
Provides utilities for deferred writing to the database over time, so as not to
cause too much interruption to the game in progress.
'''

import logging

from twisted.internet import defer, reactor

log = logging.getLogger(__name__)


DELAY = 0.01

queue = defer.DeferredQueue()
initialised = False
stopping = False


@defer.inlineCallbacks
def init():
    '''
    Initialises the deferred database queue.
    '''
    global initialised
    if initialised:
        return
    initialised = True

    reactor.addSystemEventTrigger('before', 'shutdown', teardown)

    while queue or not stopping:
        fn = yield queue.get()
        log.debug('dbqueue: running %r', fn)
        try:
            yield fn()
        except:
            log.exception('Error in queued database function')

        yield wait()


def add(fn):
    '''
    Adds the given function to the database queue.
    '''
    assert initialised
    queue.put(fn)


def wait():
    '''
    Returns a deferred which will fire when the next database command should be
    run. Database callback functions with lots of database work to do should
    call this to break up the work and avoid blocking.
    '''
    assert initialised
    d = defer.Deferred()
    if stopping:
        d.callback(None)
    else:
        reactor.callLater(DELAY, d.callback, None)
    return d


def teardown():
    assert initialised

    global stopping
    stopping = True

    d = defer.Deferred()
    log.info('Stopping database queue...')

    @queue.put
    def queueIsEmpty():
        log.info('Database queue stopped')
        d.callback(None)

    return d



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import pygame

from trosnoth.gui.framework.dialogbox import DialogBox, DialogBoxAttachedPoint
from trosnoth.gui.framework.elements import TextElement, TextButton
from trosnoth.gui.framework.utils import wrapWords
from trosnoth.utils.event import Event
from trosnoth.gui.common import Location, ScaledSize

absoluteWidth = 420
class ConnectionFailedDialog(DialogBox):
    def __init__(self, app, reason):
        width = app.screenManager.scaleFactor * absoluteWidth

        font = app.fonts.connectionFailedFont
        boundary = app.screenManager.scaleFactor * 20
        lines = wrapWords(app, reason, font, width - 2 * boundary)

        x = boundary
        elements = []
        colours = app.theme.colours
        for text in lines:
            elements.append(TextElement(app, text, font, Location(
                    DialogBoxAttachedPoint(self, (0, x), 'midtop'), 'midtop'),
                    colour=colours.errorColour))
            x += font.getHeight(app)

        # Boundary * 3 for top, bottom, and middle
        height = boundary * 3 + font.getHeight(app) * (len(lines) + 1)
        size = ScaledSize(absoluteWidth, height)
        super(ConnectionFailedDialog, self).__init__(app, size,
                "Connection Failed")

        self.elements = elements
        self.elements.append(TextButton(app, Location(DialogBoxAttachedPoint(
                self, (0, -boundary), 'midbottom'), 'midbottom'), "OK", font,
                colours.dialogButtonColour, colours.radioMouseover,
                onClick=lambda sender: self.close()))
        self.onEnter = Event()
        self.onEnter.addListener(self.close)

    def processEvent(self, event):
        if event.type == pygame.KEYDOWN and event.key in (pygame.K_KP_ENTER,
                pygame.K_RETURN):
            self.onEnter.execute()
            return None
        else:
            return super(ConnectionFailedDialog, self).processEvent(event)




# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from framework import Element
import pygame

class Hint(Element):
    reqTimeOver = 0.8
    disappearTime = 8

    def __init__(self, app, text, elementOver, font, align='left'):
        super(Hint, self).__init__(app)

        self.lines = text.split('\n')
        self.font = font
        self.align = align

        self.elementOver = elementOver

        # Check that it's valid at constructor time
        self._getOverRect()

        # Value for storing how long the mouse has been over our rect:
        self.timeOver = 0

        self.visible = False
        self.beenOff = True
        self.pos = (0,0)

    def _getOverRect(self):
        # Try to get a rect argument from elementOver
        if isinstance(self.elementOver, pygame.Rect):
            return self.elementOver
        # Try to get it from an element:
        elif hasattr(self.elementOver, '_getRect'):
            return self.elementOver._getRect()
        else:
            raise ValueError, "Not a valid elementOver argument"

    def __getSurface(self):
        bgColour = (204, 204, 0)
        rendered = []
        for line in self.lines:
            rendered.append(self.font.render(self.app, line, True, (0,0,0),
                    bgColour))
        width = 0
        height = 0
        for r in rendered:
            width = max(width, r.get_width())
            height += r.get_height()

        # Create our own image now:
        border = (3,3)
        surface = pygame.Surface((width + 2 * border[0],
                height + 2 * border[1]))
        surface.fill(bgColour)
        yPos = border[1]
        for r in rendered:
            if self.align == 'left':
                xPos = border[0]
            elif self.align == 'right':
                xPos = width - r.get_width() + border[0]
            elif self.align == 'middle':
                xPos = (width - r.get_width()) / 2 + border[0]
            else:
                raise ValueError, "Not a valid alignment argument"

            surface.blit(r, (xPos, yPos))
            yPos += r.get_height()
        pygame.draw.rect(surface, (0,0,0), surface.get_rect(), 3)
        return surface

    def processEvent(self, event):
        '''Processes the specified event and returns the event if it should
        be passed on, or None if it has been caught.'''
        # All events will make our hint invisible, unless it's a
        # mouse motion that stays within our rect, without touching
        # the hint image
        if (event.type == pygame.MOUSEMOTION and (self.visible or not
                self.beenOff)):
            rect = self.__getSurface().get_rect()
            rect.topleft = self.pos
            if (not self._getOverRect().collidepoint(event.pos)):
                # It's no longer over our rect.
                self.visible = False
                self.timeOver = 0
                self.beenOff = True

            elif rect.collidepoint(event.pos):
                    # It's on our hint! Quick! Disappear!
                    self.visible = False
                    self.timeOver = 0
        else:
            self.visible = False
            self.timeOver = 0
        return event

    def tick(self, deltaT):
        '''Gives the element a chance to update itself. deltaT is the time
        in seconds since the last tick cycle.'''
        pos = pygame.mouse.get_pos()
        if self._getOverRect().collidepoint(pos):
            self.timeOver += deltaT
            if self.timeOver > self.disappearTime:
                self.visible = False
            elif (not self.visible and self.beenOff and self.timeOver >
                    self.reqTimeOver):
                self.visible = True
                # Remember the position, and provide a little offset
                self.pos = (pos[0] + 2, pos[1] + 2)
                self.beenOff = False

    def draw(self, screen):
        '''Gives the element a chance to draw itself onto the screen.'''
        if self.visible:
            # Let's check that pos is a good position to place it - when it was
            # set in the tick method, we had no idea about the size of the
            # screen.
            surface = self.__getSurface()
            x, y = self.pos
            xChange = yChange = False
            width, height = surface.get_rect().size
            screenWidth, screenHeight = screen.get_rect().size
            if screenWidth - self.pos[0] < width:
                x = screenWidth - width
                xChange = True
            if screenHeight - self.pos[1] < height:
                y = screenHeight - height
                yChange = True
            if xChange and yChange:
                # It is extremely likely that the mouse is over
                # the hint - not convenient.
                # Let's try putting it on another side - may not be
                # successful though (i.e. if the hint is too big)
                if x > width:
                    x = x - width
                elif y > width:
                    y = y - width

            # Put the start of the text on screen
            if x < 0:
                x = 0
            if y < 0:
                y = 0

            self.pos = x, y
            screen.blit(self.__getSurface(), self.pos)



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import logging
import pickle
import socket

from trosnoth import rsa
from trosnoth.network import authcommands
from trosnoth.network.client import TrosnothClientProtocol
from trosnoth.network.networkDefines import (validServerVersions, serverVersion,
        multicastPort, multicastGroup)
from trosnoth.utils.unrepr import unrepr
from twisted.internet.error import CannotListenError
from twisted.internet import reactor, defer
from twisted.internet.protocol import ClientCreator, DatagramProtocol
from twisted.protocols import amp
from trosnoth.utils.twist import WeakCallLater

log = logging.getLogger('lobby')

class PasswordGetter(object):
    '''
    Interface for password getters, which must be used as a parameter to
    lobby.Game.join() below.
    '''
    def getPassword(self, host, errorText=''):
        '''
        Returns a deferred whose callback will be executed with (create,
        username, password), where create is a boolean indicating whether the
        account should be created. If the user cancels the getPassword(), the
        deferred's callback will be executed with an argument of None.

        @errorText: Should be provided if the previous password was invalid or
                the host refused to create a user.
        '''

class Game(object):
    '''
    Base class for lobby.Game objects.
    '''
    def join(self, passwordGetter):
        '''
        Connects to the game and returns a deferred whose callback will be
        executed with (trosnothClient, authTag).
        @param passwordGetter: an object satisfying the PasswordGetter
            interface, used if authentication is needed to join this game.
        '''

    @defer.inlineCallbacks
    def connectToGameServer(self, host, port, timeout=5):
        cc = ClientCreator(reactor, TrosnothClientProtocol)

        trosnothClient = yield cc.connectTCP(host, port, timeout=timeout)
        settings = yield trosnothClient.getSettings()

        defer.returnValue((trosnothClient, settings))


class AuthenticationCancelled(Exception):
    '''
    Raised in Game.join() if authentication was required and the user cancelled
    the password dialogue box.
    '''

class AuthGame(Game):
    def __init__(self, lobby, server, id):
        self.lobby = lobby
        self.server = server
        self.id = id

    @defer.inlineCallbacks
    def join(self, passwordGetter):
        '''
        Attempts to connect to this game. Returns a deferred whose
        callback will be executed with (trosnothClient, authTag).
        '''
        log.info('Attempting to join game on %r', self.server)
        host, port = self.server
        p = yield ClientCreator(reactor, amp.AMP).connectTCP(host, port)
        log.info('Connection established')

        try:
            result = yield p.callRemote(authcommands.ConnectToGame, id=self.id)
        except authcommands.NotAuthenticated:
            log.info('Not authenticated')
            yield authenticate(p, host, passwordGetter)

            log.info('Retrying')
            result = yield p.callRemote(authcommands.ConnectToGame, id=self.id)
        finally:
            if p.transport.connected:
                p.transport.loseConnection()

        log.info('Done')

        trosnothClient, settings  = (
            yield self.connectToGameServer(host, result['port']))
        log.info('Got trosnothClient')
        self.lobby.app.identitySettings.setNick(result['nick'])
        defer.returnValue((trosnothClient, settings, result['authTag']))

class NonAuthGame(Game):
    def __init__(self, lobby, ipAddr, port):
        self.lobby = lobby
        self.ip = ipAddr
        self.port = port

    @defer.inlineCallbacks
    def join(self, passwordGetter):
        '''
        Attempts to connect to the given lobby.Game. Returns a deferred whose
        callback will be executed with (trosnothClient, authTag).
        '''
        trosnothClient, settings = (
            yield self.connectToGameServer(self.ip, self.port))
        defer.returnValue((trosnothClient, settings, 0))

class IncorrectServerVersion(Exception):
    '''
    The server created a game of the wrong version.
    '''

class AuthServerConnection(object):
    def __init__(self, server, timeout=5):
        self.host, self.port = server
        self.timeout = timeout
        self.protocol = None

    def __enter__(self):
        d = ClientCreator(reactor, amp.AMP).connectTCP(self.host, self.port,
                timeout=self.timeout)
        d.addCallback(self.connectionEstablished)
        return d

    def connectionEstablished(self, protocol):
        self.protocol = protocol
        return protocol

    def __exit__(self, type, value, traceback):
        if self.protocol and self.protocol.transport.connected:
            self.protocol.transport.loseConnection()

class Lobby(object):
    '''
    Performs tasks related to finding games and joining them.
    '''
    def __init__(self, app):
        self.app = app

    @defer.inlineCallbacks
    def getGames(self, server, timeout=2):
        '''
        Connects to an authentication server and asks it for the available
        games on that server. Returns a deferred whose callback will be executed
        with a list of lobby.Game objects.
        @param server: A tuple of (host, port).
        '''
        with AuthServerConnection(server, timeout) as deferred:
            p = yield deferred

            result = yield p.callRemote(authcommands.ListGames)
            games = []
            for item in result['games']:
                if (item['game'] == 'Trosnoth1' and item['version'] in
                        validServerVersions):
                    games.append(AuthGame(self, server, item['id']))

            defer.returnValue(games)

    @defer.inlineCallbacks
    def getOtherGames(self, server, timeout=2):
        '''
        Connects to an authentication server and asks it for the available
        games it knows about which are not hosted on that server. Returns a
        deferred whose callback will be executed with a list of lobby.Game
        objects.
        @param server: A tuple of (host, port).
        '''
        with AuthServerConnection(server, timeout) as deferred:
            p = yield deferred

            result = yield p.callRemote(authcommands.ListOtherGames)
            games = []
            for item in result['games']:
                if (item['game'] == 'Trosnoth1' and item['version'] in
                        validServerVersions):
                    games.append(NonAuthGame(self, item['ip'], item['port']))

            defer.returnValue(games)

    @defer.inlineCallbacks
    def getMulticastGames(self):
        '''
        Returns a deferred which will be called back with a sequence of Game
        objects representing the games which could be found on the LAN by
        multicasting a game request.
        '''
        games = yield getMulticastGames()

        result = []
        for (ip, port), gameInfo in games:
            result.append(NonAuthGame(self, ip, port))
        defer.returnValue(result)

    @defer.inlineCallbacks
    def registerGame(self, authServer, gameServer, timeout=2):
        '''
        Registers the game running on the given TrosnothServerFactory with the
        given authentication server. Returns a deferred whose callback will be
        executed with True or False indicating whether the registration has
        been successful.
        @param authServer: A tuple of (host, port).
        '''
        with AuthServerConnection(authServer, timeout) as deferred:
            p = yield deferred

            try:
                yield p.callRemote(authcommands.RegisterGame,
                        game='Trosnoth1', version=serverVersion,
                        port=gameServer.getTCPPort())
            except:
                defer.returnValue(False)
            else:
                defer.returnValue(True)

    @defer.inlineCallbacks
    def startGame(self, server, passwordGetter):
        '''
        Attempts to start a game on the given server, and connects to the
        newly-created game. Returns a deferred whose callback will be executed
        with (trosnothClient, authTag).
        @param server: A tuple of (host, port).
        @param passwordGetter: PasswordGetter object which is used to
            authenticate with the server if needed.
        '''
        with AuthServerConnection(server) as deferred:
            p = yield deferred

            result = yield p.callRemote(authcommands.CreateGame,
                    game='Trosnoth1')
            if result['version'] not in validServerVersions:
                raise IncorrectServerVersion()

            game = AuthGame(self, server, result['id'])

            result = yield game.join(passwordGetter)
            defer.returnValue(result)

@defer.inlineCallbacks
def authenticate(p, host, passwordGetter):
    errorMsg = ''
    for i in xrange(15):
        if host in AUTHENTICATION_CACHE:
            log.info('Got cached auth info')
            create = False
            username, password = AUTHENTICATION_CACHE[host]
        else:
            log.info('Checking with password getter')
            result = yield passwordGetter.getPassword(host, errorMsg)
            if result is None:
                raise AuthenticationCancelled()
            create, username, password = result

        data = yield encryptPassword(p, password)

        if create:
            log.info('Creating user')
            result = yield p.callRemote(
                    authcommands.CreateUserWithPassword,
                    username=str(username), password=data)
            result = result['result']
            if result == '':
                log.info('Auth successful')
                return  # Authentication successful.
            errorMsg = result
        else:
            log.info('Attempting to auth')
            result = yield p.callRemote(authcommands.PasswordAuthenticate,
                    username=str(username), password=data)
            result = result['result']
            if result:
                log.info('Auth successful')
                return  # Authentication successful.
            errorMsg = 'Bad username/password'

    # If we can't get the authentication right in 15 tries, something is
    # wrong.
    raise AuthenticationCancelled()

@defer.inlineCallbacks
def encryptPassword(p, password):
    log.info('Getting auth token')
    token = yield p.callRemote(authcommands.GetAuthToken)
    token = token['token']

    log.info('Getting server public key')
    pubkey = yield p.callRemote(authcommands.GetPublicKey)
    pubkey = makePublicKey(pubkey)

    defer.returnValue(rsa.encrypt(token + password.encode(), pubkey))

AUTHENTICATION_CACHE = {}   # host -> (username, password)

def makePublicKey(result):
    '''
    Takes the result of the GetPublicKey amp call and converts it into a public
    key which the rsa module can handle.
    '''
    return result
    return {
        'e': pickle.decode_long(result['e']),
        'n': pickle.decode_long(result['n']),
    }

class UDPMulticaster(object):
    def __init__(self, gameGetter):
        self.stopped = False
        self.port = None
        self.listener = UDPMulticastListener(gameGetter)
        self.tryListening()

    def stop(self):
        if self.port is not None:
            self.port.stopListening()
        self.stopped = True

    def tryListening(self):
        if self.stopped:
            return
        try:
            self.port = reactor.listenMulticast(multicastPort, self.listener)
        except CannotListenError:
            # Cannot listen to the multicast, possibly because another
            # instance is running on this computer.

            # Try listening again in 5 seconds.
            WeakCallLater(5, self, 'tryListening')

MULTICAST_PROTOCOL_NAME = 'TrosnothMulticastProtocol1'

class UDPMulticastListener(DatagramProtocol):
    def __init__(self, gameGetter):
        '''
        gameGetter must be callable, and return a sequence of game information
        dicts, each of which must contain:
            ['port'] - the port this game is listening on
            ['version'] - the version string
            ['name'] - the game name
        '''
        self.getGames = gameGetter

    def startProtocol(self):
        # Join the correct multicast group.
        self.transport.joinGroup(multicastGroup)

    def datagramReceived(self, datagram, address):
        '''
        A multicast datagram has been received.
        '''
        if datagram == '%s:GetGames' % (MULTICAST_PROTOCOL_NAME,):
            for game in self.getGames():
                self.transport.write('%s:Game:%s' % (MULTICAST_PROTOCOL_NAME,
                        repr(game),), address)

class UDPMulticastGameGetter(DatagramProtocol):
    def __init__(self):
        self.port = reactor.listenUDP(0, self)
        self._deferred = None
        self._games = []

    def stop(self):
        self.port.stopListening()

    def getGames(self, timeout=1):
        '''
        Returns a deferred which will be called back with a sequence of
        (address, info) where info is a dict which has:
            ['name'] - the name of the game
            ['version'] - the version string of the game
        '''
        d= defer.Deferred()
        try:
            self.transport.write('%s:GetGames' % MULTICAST_PROTOCOL_NAME,
                    (multicastGroup, multicastPort))
        except socket.error, e:
            log.info('Could not request games from multicast: %s', e)
            d.callback([])
            return d

        self._deferred = d
        self._games = []

        WeakCallLater(timeout, self, '_gotGames', d)

        return d

    def _gotGames(self, d):
        d.callback(self._games)
        self._deferred = None
        self._games = []

    def datagramReceived(self, datagram, address):
        '''
        A reply to our query has been received.
        '''
        if self._deferred is None:
            return
        if datagram.startswith('%s:Game:' % (MULTICAST_PROTOCOL_NAME,)):
            gameInfo = unrepr(datagram[len('%s:Game:' %
                    (MULTICAST_PROTOCOL_NAME,)):])
            gameAddress = (address[0], gameInfo.pop('port'))
            self._games.append((gameAddress, gameInfo))

def getMulticastGames(timeout=1):
    getter = UDPMulticastGameGetter()
    d = getter.getGames(timeout)
    @d.addCallback
    def gotGames(games):
        getter.stop()
        return games
    return d



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import platform
import webbrowser

import pygame

def openPage(app, url):
    # If we're in fullscreen, minimise.
    if app.screenManager.isFullScreen():
        if platform.system() != 'Darwin':
            # Try minimising.
            pygame.display.iconify()
        else:
            # .iconify() has problems on Mac OSX

            from trosnoth.gui.framework.dialogbox import OkBox
            from trosnoth.gui.common import ScaledSize
            # Switch out of full screen.
            app.displaySettings.fullScreen = False
            app.displaySettings.apply()
            box = OkBox(app, ScaledSize(600, 300),
                'Web browser opened',
                'Please refer to web browser.'
            )
            def revert():
                # Switch back to full screen.
                app.displaySettings.fullScreen = True
                app.displaySettings.apply()
            box.onClose.addListener(revert)
            box.show()

    # Open in a new tab if possible.
    try:
        webbrowser.open_new_tab(url)
    except AttributeError:
        webbrowser.open_new(url)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='AchievementProgress',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('achievementId', models.TextField()),
                ('unlocked', models.BooleanField(default=False)),
                ('progress', models.IntegerField(null=True, blank=True)),
                ('data', models.TextField(blank=True)),
            ],
            options={
                'verbose_name_plural': 'Achievement progress records',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='GamePlayer',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('bot', models.BooleanField(default=False)),
                ('team', models.CharField(max_length=1)),
                ('upgradesUsed', models.TextField(default=b'', blank=True)),
                ('starsEarned', models.IntegerField(default=0)),
                ('starsWasted', models.IntegerField(default=0)),
                ('starsUsed', models.IntegerField(default=0)),
                ('kills', models.IntegerField(default=0)),
                ('deaths', models.IntegerField(default=0)),
                ('zoneTags', models.IntegerField(default=0)),
                ('zoneAssists', models.IntegerField(default=0)),
                ('shotsFired', models.IntegerField(default=0)),
                ('shotsHit', models.IntegerField(default=0)),
                ('timeAlive', models.FloatField(default=0)),
                ('timeDead', models.FloatField(default=0)),
                ('killStreak', models.IntegerField(default=0)),
                ('tagStreak', models.IntegerField(default=0)),
                ('aliveStreak', models.FloatField(default=0)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='GameRecord',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('started', models.DateTimeField()),
                ('finished', models.DateTimeField()),
                ('gameSeconds', models.FloatField(default=0)),
                ('mapHeight', models.IntegerField()),
                ('halfMapWidth', models.IntegerField()),
                ('serverVersion', models.TextField()),
                ('blueTeamName', models.TextField()),
                ('redTeamName', models.TextField()),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='PlayerKills',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('count', models.IntegerField(default=0)),
                ('game', models.ForeignKey(to='trosnoth.GameRecord')),
            ],
            options={
                'verbose_name_plural': 'Player kills records',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='TrosnothServerSettings',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('serverName', models.TextField(default=b'My First Trosnoth Server')),
                ('welcomeText', models.TextField(default=b'Congratulations! You have successfully installed your Trosnoth server. <a href="admin/trosnoth/trosnothserversettings/">Click here</a> to configure it.')),
            ],
            options={
                'verbose_name_plural': 'Trosnoth server setting',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='TrosnothUser',
            fields=[
                ('user', models.OneToOneField(primary_key=True, serialize=False, to=settings.AUTH_USER_MODEL)),
                ('nick', models.TextField(unique=True)),
                ('lastSeen', models.DateTimeField(null=True, blank=True)),
                ('oldPasswordHash', models.BinaryField(default=b'')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='playerkills',
            name='killee',
            field=models.ForeignKey(related_name='+', to='trosnoth.TrosnothUser'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='playerkills',
            name='killer',
            field=models.ForeignKey(related_name='+', to='trosnoth.TrosnothUser'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='gameplayer',
            name='game',
            field=models.ForeignKey(to='trosnoth.GameRecord'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='gameplayer',
            name='user',
            field=models.ForeignKey(to='trosnoth.TrosnothUser'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='achievementprogress',
            name='user',
            field=models.ForeignKey(to='trosnoth.TrosnothUser'),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('trosnoth', '0004_auto_20160208_2327'),
    ]

    operations = [
        migrations.AlterField(
            model_name='trosnothserversettings',
            name='iceProxyStringOverride',
            field=models.TextField(default=b'', blank=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='trosnothserversettings',
            name='iceSecret',
            field=models.TextField(default=b'', blank=True),
            preserve_default=True,
        ),
    ]



import logging

from trosnoth.const import TICK_PERIOD
from trosnoth.model.hub import Hub, Node
from trosnoth.messages import TickMsg
from trosnoth.utils.twist import WeakLoopingCall
from trosnoth.utils.utils import timeNow

log = logging.getLogger(__name__)


class UIMsgThrottler(Hub, Node):
    '''
    This class is used for RemoteGames which are connected to the UI. It delays
    messages a little bit, in order that the UI can smoothly display what is
    happening even if there is some variation in network round trip time.
    '''

    def __init__(self, *args, **kwargs):
        super(UIMsgThrottler, self).__init__(*args, **kwargs)
        self.lagAnalyser = LagAnalyser()
        self.agentEventGroups = [[]]
        self._tweenFraction = 1
        self._uiCatchUp = 0

    def stop(self):
        pass

    def disconnectNode(self):
        super(UIMsgThrottler, self).disconnectNode()
        self.hub.disconnectNode()

    def connectNewAgent(self):
        return self.hub.connectNewAgent()

    def disconnectAgent(self, agentId):
        if self.hub:
            self.hub.disconnectAgent(agentId)

    def sendRequestToGame(self, agentId, msg):
        if self.hub:
            self.hub.sendRequestToGame(agentId, msg)

    def gotServerCommand(self, msg):
        if isinstance(msg, TickMsg):
            self.agentEventGroups[-1].append((self.node.gotServerCommand, msg))
            self.agentEventGroups.append([])

            self.lagAnalyser.gotTick()
            idealLag = self.lagAnalyser.getIdealUILag()
            tweenFraction = self._tweenFraction
            currentLag = (
                len(self.agentEventGroups) - tweenFraction) * TICK_PERIOD
            self._uiCatchUp = currentLag - idealLag
        elif msg.isControl:
            self.node.gotServerCommand(msg)
        else:
            self.agentEventGroups[-1].append((self.node.gotServerCommand, msg))

    def gotMessageToAgent(self, agentId, msg):
        if msg.isControl:
            self.node.gotMessageToAgent(agentId, msg)
        else:
            self.agentEventGroups[-1].append((
                self.node.gotMessageToAgent, agentId, msg))

    def agentDisconnected(self, agentId):
        self.node.agentDisconnected(agentId)

    def releaseEventGroup(self):
        if len(self.agentEventGroups) < 2:
            self._tweenFraction = 1
            return
        group = self.agentEventGroups.pop(0)
        for event in group:
            event[0](*event[1:])
        self._tweenFraction = max(0, self._tweenFraction - 1)

    def uiTick(self, deltaT):
        '''
        Called to indicate that the UI has advanced by the given time delta.
        Returns the tween fraction to use in the UniverseGUI.
        '''
        origDeltaT = deltaT
        catchUp = self._uiCatchUp - deltaT
        deltaT = max(0, deltaT + catchUp * (1 - 0.1 ** deltaT))

        self._uiCatchUp -= (deltaT - origDeltaT)
        self._tweenFraction += deltaT / TICK_PERIOD
        while self._tweenFraction > 1:
            # If the frame rate is really low, we may have to tick more than
            # one frame at a time.
            self.releaseEventGroup()
        return self._tweenFraction


class LocalGameTweener(object):
    '''
    Used in place of a UIMsgThrottler for a local game to simply get the
    current tween fraction.
    '''
    def __init__(self, game):
        self.game = game
        self.tweenFraction = 0
        game.onServerCommand.addListener(self.gotServerCommand)

    def stop(self):
        self.game.onServerCommand.removeListener(self.gotServerCommand)

    def gotServerCommand(self, msg):
        if isinstance(msg, TickMsg):
            self.tweenFraction = 0

    def uiTick(self, deltaT):
        self.tweenFraction += deltaT / TICK_PERIOD
        self.tweenFraction = min(1, self.tweenFraction)
        return self.tweenFraction


class LagAnalyser(object):
    TIME_SPAN = 5
    NOTE_TIME = 0.1

    def __init__(self):
        self.lastNote = None
        self.ignoreUntil = None
        self.reset()
        self.monitor = WeakLoopingCall(self, 'notePeriod')
        self.monitor.start(self.NOTE_TIME)

    def reset(self):
        self.startTime = timeNow()
        self.ticks = [self.startTime]

    def notePeriod(self):
        '''
        Called periodically to allow the analyser to notice if the reactor is
        not running smoothly.
        '''
        now = timeNow()
        if self.lastNote is None:
            self.lastNote = now
            return
        period = now - self.lastNote
        self.lastNote = now

        if period - self.NOTE_TIME >= self.NOTE_TIME:
            # Reactor congestion: reset tally
            self.ignoreUntil = now + self.NOTE_TIME

    def gotTick(self):
        now = timeNow()

        if self.ignoreUntil is not None:
            if now > self.ignoreUntil:
                self.reset()
            else:
                return

        self.ticks.append(now)
        limit = now - self.TIME_SPAN
        while self.ticks[0] < limit:
            self.ticks.pop(0)

    def getIdealUILag(self):
        '''
        Based on the ticks received, calculates the time which the UI should
        lag behind receiving messages, in order for things to be displayed
        smoothly.
        '''
        final = self.ticks[-1]
        initial = max(self.startTime, final - self.TIME_SPAN)

        oldT = initial
        oldValue = upper = lower = 0
        for t in self.ticks:
            value = oldValue + (t - oldT)
            upper = max(upper, value)
            value -= TICK_PERIOD
            lower = min(lower, value)
            oldT, oldValue = t, value

        return upper - lower



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import heapq
import os
import re
import subprocess
import sys

EXCLUDE_DIRS = set(['test'])
POOR_FUNCTION_RE = re.compile('\s*def [A-Z]')
IMPORT_STAR_RE = re.compile('from .+ import \*')

THIS_FILE = os.path.abspath(__file__)


def main():
    if len(sys.argv) > 2:
        print 'Expected one arg only.'
        sys.exit(1)

    if len(sys.argv) == 2:
        path = sys.argv[1]
        if os.path.isfile(path):
            analyseFile(path)
            return
    else:
        path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))

    results = []
    for base, dirs, files in os.walk(path):
        dirs[:] = [
            d for d in dirs if not d.startswith('.') and d not in EXCLUDE_DIRS]
        for f in files:
            if f.endswith('.py'):
                filename = os.path.join(base, f)
                s = score(filename)
                if s:
                    heapq.heappush(results, (-s, filename))

    for i in xrange(20):
        if len(results) == 0:
            print '--- no more bad files ---'
            break
        print '%d: %s' % heapq.heappop(results)


def score(filename):
    if filename == THIS_FILE:
        return 0

    contents = [
        l for l in open(filename, 'rb').read().splitlines()
        if 'pragma: noqa' not in l]
    return (
        longLineCount(contents) +
        printCount(contents) +
        trailingWhitespaceCount(contents) +
        trailingBackslashCount(contents) +
        5 * tabCount(contents) +
        3 * pyflakesScore(filename) +
        2 * todoCount(contents) +
        2 * poorlyNamedFunctionCount(contents) +
        (25 if containsImportStar(contents) else 0) +
        (15 if importsPygameLocals(contents) else 0)
    )


def analyseFile(filename):
    contents = [
        l for l in open(filename, 'rb').read().splitlines()
        if 'pragma: noqa' not in l]
    print 'Long lines: %d' % (longLineCount(contents),)
    print 'prints: %d' % (printCount(contents),)
    print 'trailing whitespaces: %d' % (trailingWhitespaceCount(contents),)
    print 'trailing backslashes: %d' % (trailingBackslashCount(contents),)
    print 'tabs: %d' % (tabCount(contents),)
    print 'pyflakes score: %d' % (pyflakesScore(filename),)
    print 'TODO lines: %d' % (todoCount(contents),)
    print 'poorly named functions: %d' % (poorlyNamedFunctionCount(contents),)
    print 'contains import *: %r' % (containsImportStar(contents),)
    print 'imports pygame.locals: %r' % (importsPygameLocals(contents),)


def longLineCount(contents):
    '''
    Returns the number of lines in the file which are longer than 80 characters
    wide.
    '''
    return sum([1 if len(l) > 80 else 0 for l in contents])


def trailingBackslashCount(contents):
    '''
    Returns the number of lines in the file which are longer than 80 characters
    wide.
    '''
    return sum([1 if l.rstrip().endswith('\\') else 0 for l in contents])


def tabCount(contents):
    '''
    Returns the number of lines in the file which contain tabs.
    '''
    return sum([1 if '\t' in l else 0 for l in contents])


def printCount(contents):
    '''
    Returns the number of print statements apparently in the code.
    '''
    return sum([1 if l.lstrip().startswith('print ') else 0 for l in contents])


def todoCount(contents):
    '''
    Returns the number of lines which contain TODO, FIXME or XXX.
    '''
    return sum([
        1 if ('TODO' in l or 'FIXME' in l or 'XXX' in l) else 0
        for l in contents])


def poorlyNamedFunctionCount(contents):
    '''
    Returns the number of functions with capitalised names.
    '''
    return sum([1 if POOR_FUNCTION_RE.match(l) else 0 for l in contents])


def trailingWhitespaceCount(contents):
    '''
    Returns the number of lines in the file which have trailing whitespace.
    '''
    return sum([1 if l.rstrip() != l else 0 for l in contents])


def containsImportStar(contents):
    '''
    Returns True or False depending on whether the file contains import *.
    '''
    return any(IMPORT_STAR_RE.match(l) for l in contents)


def pyflakesScore(filename):
    cmd = ['pyflakes', filename]
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    p.communicate()
    return p.returncode


def importsPygameLocals(contents):
    '''
    Returns True if pygame.locals is imported.
    '''
    for line in contents:
        if (line.startswith('from pygame import locals') or
                line.startswith('import pygame.locals')):
            return True
    return False

if __name__ == '__main__':
    main()



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from trosnoth.gui.framework.elements import TextElement, TextButton
from trosnoth.gui.framework import framework
from trosnoth.gui.common import ScaledLocation
from trosnoth.gui.common import (Location, ScaledScreenAttachedPoint,
        ScaledSize)

class ConnectingScreen(framework.CompoundElement):
    def __init__(self, app, serverName = 'server', onCancel=None):
        super(ConnectingScreen, self).__init__(app)
        colours = app.theme.colours

        self.text = TextElement(self.app, 'Connecting to %s...' % serverName,
                self.app.screenManager.fonts.bigMenuFont,
                ScaledLocation(512, 384, 'center'),
                colour = colours.connectingColour)

        button = TextButton(
            self.app,
            Location(ScaledScreenAttachedPoint(ScaledSize(0, 300),
                    'center'), 'center'),
            'cancel',
            self.app.screenManager.fonts.bigMenuFont,
            colours.mainMenuColour,
            colours.white,
            onClick=onCancel
        )
        self.onCancel = button.onClick

        self.elements = [button, self.text]

    def setServer(self, serverName):
        self.text.setText('Connecting to %s...' % serverName)
        self.text.setFont(self.app.screenManager.fonts.bigMenuFont)



import logging

from trosnoth.const import (
    PRIVATE_CHAT, TEAM_CHAT, OPEN_CHAT, RESYNC_IN_PROGRESS_REASON,
    GAME_NOT_STARTED_REASON, ALREADY_ALIVE_REASON, BE_PATIENT_REASON,
    ENEMY_ZONE_REASON, FROZEN_ZONE_REASON,
)
from trosnoth.messages.base import (
    AgentRequest, ServerCommand, ServerResponse, ClientCommand,
)
from trosnoth.utils import globaldebug

log = logging.getLogger(__name__)


class TickMsg(ServerCommand):
    idString = 'tick'
    fields = ('tickId',)
    packspec = 'H'


class TaggingZoneMsg(ServerCommand):
    idString = 'Tag!'
    fields = 'zoneId', 'playerId', 'teamId'
    packspec = 'Icc'


class CreateCollectableStarMsg(ServerCommand):
    idString = 'StCr'
    fields = 'starId', 'xPos', 'yPos'
    packspec = 'cff'


class RemoveCollectableStarMsg(ServerCommand):
    idString = 'StCo'
    fields = 'starId'
    packspec = 'c'


class BasePlayerUpdate(ServerCommand):
    '''
    attached may be:
        'f' - falling
        'g' - on ground
        'l' - on wall to left of player
        'r' - on wall to right of player
    '''
    fields = (
        'playerId', 'xPos', 'yPos', 'yVel', 'angle', 'ghostThrust',
        'jumpTime', 'gunReload', 'respawn', 'attached', 'stars', 'health',
        'resync', 'leftKey', 'rightKey', 'jumpKey', 'downKey')
    packspec = 'cffffffffcBB?????'


class PlayerUpdateMsg(BasePlayerUpdate):
    idString = 'PlUp'

    def applyOrderToWorld(self, world):
        player = world.getPlayer(self.playerId)
        player.applyPlayerUpdate(self)


class ResyncPlayerMsg(BasePlayerUpdate):
    '''
    Carries the same information as PlayerUpdateMsg, but is sent only to the
    player in question to tell them to reposition their player and acknowledge
    the resync.
    '''
    idString = 'Sync'
    isControl = True


class DelayUpdatedMsg(ServerResponse):
    idString = 'Dlay'
    fields = 'delay'
    packspec = 'f'

    def applyOrderToLocalState(self, localState, world):
        localState.serverDelay = self.delay


class CheckSyncMsg(AgentRequest):
    '''
    Sent periodically by the client to confirm that it's idea of where its
    player is matches the server's idea.
    '''
    idString = 'syn?'
    fields = 'tickId', 'xPos', 'yPos', 'yVel'
    packspec = 'Hfff'
    timestampedPlayerRequest = True

    def serverApply(self, game, agent):
        if agent.player is None or agent.player.resyncing:
            return

        x, y = agent.player.pos
        if (
                abs(x - self.xPos) >= 1
                or abs(y - self.yPos) >= 1
                or abs(agent.player.yVel - self.yVel) >= 1):
            log.info('Player out of sync: %s', agent.player)
            agent.player.sendResync()


class ResyncAcknowledgedMsg(ClientCommand):
    idString = 'Synd'
    fields = (
        'tickId', 'xPos', 'yPos', 'yVel', 'angle', 'ghostThrust', 'playerId')
    packspec = 'Hfffffc'
    timestampedPlayerRequest = True
    playerId = '\x00'

    def serverApply(self, game, agent):
        if agent.player is None:
            return
        if not agent.player.resyncing:
            log.warning(
                'Received unexpected resync acknowledgement from %s',
                agent.player)
            # If we get a resync acknowledgement when we're not resyncing
            # something is up and we need to resync again.
            agent.player.sendResync()
            return

        if not agent.player.checkResyncAcknowledgement(self):
            # If the values the client is acknowledging don't match up to where
            # we think the client is at, it's probably because another resync /
            # a game reset has happened, so we need to wait for the second
            # acknowledgement.
            return

        self.playerId = agent.player.id
        game.sendServerCommand(self)

    def applyRequestToLocalState(self, localState):
        localState.player.resyncing = False

    def applyOrderToWorld(self, world):
        player = world.getPlayer(self.playerId)
        player.resyncing = False


class RespawnMsg(ServerCommand):
    idString = 'Resp'
    fields = 'playerId', 'zoneId'
    packspec = 'cI'


class RespawnRequestMsg(AgentRequest):
    idString = 'Resp'
    fields = 'tickId'
    packspec = 'H'
    timestampedPlayerRequest = True

    def clientValidate(self, localState, world, sendResponse):
        if localState.player is None:
            return False
        code = self.getInvalidCode(localState.player, world)
        if code is None:
            return True

        sendResponse(CannotRespawnMsg(code))
        return False

    def applyRequestToLocalState(self, localState):
        localState.player.respawn()

    def serverApply(self, game, agent):
        # Since the client has already applied the respawn before the message
        # gets to the server, any invalid respawn necessitates a resync.
        if not agent.player or agent.player.resyncing:
            return
        code = self.getInvalidCode(agent.player, game.world)
        if code is None:
            game.sendServerCommand(
                RespawnMsg(agent.player.id, agent.player.getZone().id))
        else:
            if code == ENEMY_ZONE_REASON:
                reason = 'Sorry, too slow! You no longer own the zone.'
            else:
                reason = 'Your computer was out of sync with the server!'
            agent.player.sendResync(reason=reason)

    def getInvalidCode(self, player, world):
        '''
        Returns None if this respawn request seems valid, a reason code
        otherwise.
        '''
        if player.resyncing:
            return RESYNC_IN_PROGRESS_REASON
        if not world.abilities.respawn:
            return GAME_NOT_STARTED_REASON
        if not player.dead:
            return ALREADY_ALIVE_REASON
        if player.respawnGauge > 0:
            return BE_PATIENT_REASON
        if not player.inRespawnableZone():
            return ENEMY_ZONE_REASON
        if player.getZone().frozen:
            return FROZEN_ZONE_REASON
        return None


class CannotRespawnMsg(ServerResponse):
    '''
    reasonId may be:
        P: game hasn't started
        A: Already Alive
        T: Can't respawn yet
        E: In enemy zone
        F: Frozen Zone

    Note that this message actually always originates on the client rather than
    the server, so it never needs to travel on the network.
    '''
    idString = 'NoRs'
    fields = 'reasonId'
    packspec = 'c'


class UpdatePlayerStateMsg(ClientCommand):
    idString = 'Pres'
    fields = 'value', 'tickId', 'playerId', 'stateKey'
    packspec = 'bHc*'
    timestampedPlayerRequest = True
    playerId = '\x00'

    def clientValidate(self, localState, world, sendResponse):
        if not localState.player:
            return False
        return True

    def applyRequestToLocalState(self, localState):
        localState.player.updateState(self.stateKey, self.value)

    def serverApply(self, game, agent):
        if not agent.player or agent.player.resyncing:
            return
        self.playerId = agent.player.id
        game.sendServerCommand(self)

    def applyOrderToWorld(self, world):
        player = world.getPlayer(self.playerId)
        if player:
            player.updateState(self.stateKey, self.value)


class AimPlayerAtMsg(ClientCommand):
    idString = 'Aim@'
    fields = 'angle', 'thrust', 'tickId', 'playerId'
    packspec = 'ffHc'
    timestampedPlayerRequest = True
    playerId = '\x00'

    def applyRequestToLocalState(self, localState):
        localState.player.lookAt(self.angle, self.thrust)

    def serverApply(self, game, agent):
        if agent.player is None or agent.player.resyncing:
            return
        self.playerId = agent.player.id
        game.sendServerCommand(self)

    def applyOrderToWorld(self, world):
        player = world.getPlayer(self.playerId)
        if player:
            player.lookAt(self.angle, self.thrust)


class ShootMsg(AgentRequest):
    idString = 'shot'
    fields = 'tickId', 'localId'
    packspec = 'HH'
    localId = 0
    timestampedPlayerRequest = True

    def clientValidate(self, localState, world, sendResponse):
        if not world.canShoot():
            return False
        if not localState.player.canShoot():
            return False
        return True

    def applyRequestToLocalState(self, localState):
        self.localId = localState.shotFired()

    def serverApply(self, game, agent):
        if not game.world.canShoot():
            return
        if agent.player is None or agent.player.resyncing:
            return
        if not agent.player.canShoot():
            return
        if __debug__ and globaldebug.enabled:
            if (
                    globaldebug.shotLimit and
                    len(game.world.shotWithId) >= globaldebug.shotLimit):
                # When debugging shot physics, limit number of shots
                return

        xpos, ypos = agent.player.pos
        if agent.player.shoxwave:
            game.sendServerCommand(
                FireShoxwaveMsg(agent.player.id, xpos, ypos))
        else:
            shotId = game.idManager.newShotId()
            if shotId is not None:
                game.sendServerCommand(
                    ShotFiredMsg(agent.player.id, shotId, self.localId))


class ShotFiredMsg(ServerCommand):
    '''
    id manager -> agents
    '''
    idString = 'SHOT'
    fields = 'playerId', 'shotId', 'localId'
    packspec = 'cIH'

    def applyOrderToLocalState(self, localState, world):
        if localState.player and self.playerId == localState.player.id:
            localState.matchShot(self.localId, self.shotId)


class FireShoxwaveMsg(ServerCommand):
    '''
    id manager -> agents
    '''
    idString = 'Shox'
    fields = 'playerId', 'xpos', 'ypos'
    packspec = 'cff'


class ShotHitPlayerMsg(ServerCommand):
    '''
    The only reason shot collision is calculated server-side is that there are
    some times when a single shot may hit multiple players in the one tick. In
    that case, in order for the selection of killed player to be consistent,
    the server decides and broadcasts the result.
    '''
    idString = 'HitP'
    fields = 'playerId', 'shotId'
    packspec = 'cI'

    def applyOrderToWorld(self, world):
        shot = world.getShot(self.shotId)
        player = world.getPlayer(self.playerId)
        shot.collidedWithPlayer(player)


class PlayerNoticedZombieHitMsg(ServerCommand):
    '''
    Sent from the server at the point in time when a given player noticed that
    it had been hit.

    Valid values for deathType are defined in trosnoth.const.
    '''
    idString = 'Ouch'
    fields = 'playerId', 'killerId', 'deathType'
    packspec = 'ccc'

    def serverApply(self, game, agent):
        # In order for this message to enter the system at the correct time, it
        # acts as though it comes from the agent, even though it's really
        # generated by the AgentInfo object. This is why we need to implement
        # serverApply().
        game.sendServerCommand(self)

    def applyOrderToWorld(self, world):
        player = world.getPlayer(self.playerId)
        killer = world.getPlayer(self.killerId)
        player.noticeZombieHit(killer, self.deathType)


####################
# Communication
####################

class ChatFromServerMsg(ServerCommand):
    idString = 'ahoy'
    fields = 'error', 'text'
    packspec = '?*'


class ChatMsg(ClientCommand):
    '''
    Valid values for kind are defined in trosnoth.const.
    '''
    idString = 'chat'
    fields = 'kind', 'targetId', 'playerId', 'text'
    packspec = 'ccc*'
    playerId = '\x00'
    targetId = '\x00'

    def serverApply(self, game, agent):
        if agent.player is None:
            return

        if self.kind == TEAM_CHAT:
            if not agent.player.isFriendsWithTeam(
                    game.world.getTeam(self.targetId)):
                # Cannot send to opposing team.
                return
        elif self.kind == PRIVATE_CHAT:
            if game.world.getPlayer(self.targetId) is None:
                return
        elif self.kind != OPEN_CHAT:
            log.warning('unknown chat kind: %r', self.kind)
            return

        self.playerId = agent.player.id
        game.sendServerCommand(self)

    def applyOrderToWorld(self, world):
        sender = world.getPlayer(self.playerId)
        text = self.text.decode()
        if self.kind == OPEN_CHAT:
            world.onOpenChatReceived(text, sender)
        elif self.kind == PRIVATE_CHAT:
            target = world.getPlayer(self.targetId)
            target.onPrivateChatReceived(text, sender)
        elif self.kind == TEAM_CHAT:
            team = world.getTeam(self.targetId)
            world.onTeamChatReceived(team, text, sender)

    def applyOrderToLocalState(self, localState, world):
        if self.kind == PRIVATE_CHAT:
            player = localState.player
            if player and player.id == self.targetId:
                sender = world.getPlayer(self.playerId)
                text = self.text.decode()
                player.onPrivateChatReceived(text, sender)



# This file is in the public domain
import compiler


def getObj(s):
    s = "a=" + s
    return compiler.parse(s).getChildren()[1].getChildren()[0].getChildren()[1]


class UnknownType(Exception):
    pass


class Builder:

    def build(self, o):
        m = getattr(self, 'build_'+o.__class__.__name__, None)
        if m is None:
            raise UnknownType(o.__class__.__name__)
        return m(o)

    def build_List(self, o):
        return map(self.build, o.getChildren())

    def build_Const(self, o):
        return o.value

    def build_Dict(self, o):
        d = {}
        i = iter(map(self.build, o.getChildren()))
        for el in i:
            d[el] = i.next()
        return d

    def build_Tuple(self, o):
        return tuple(self.build_List(o))

    def build_Name(self, o):
        if o.name == 'None':
            return None
        if o.name == 'True':
            return True
        if o.name == 'False':
            return False
        raise UnknownType('Name')

    def build_Add(self, o):
        real, imag = map(self.build_Const, o.getChildren())
        try:
            real = float(real)
        except TypeError:
            raise UnknownType('Add')
        if not isinstance(imag, complex) or imag.real != 0.0:
            raise UnknownType('Add')
        return real+imag

    def build_UnarySub(self, o):
        (child,) = o.getChildren()
        return -self.build(child)


def unrepr(s):
    return Builder().build(getObj(s))



from math import atan2, sin, cos, pi, floor
import logging

from trosnoth.const import PATH_FINDING_DISCRETISATION_UNIT as D_UNIT
from trosnoth.utils import math

log = logging.getLogger(__name__)


class Obstaclish(object):
    ground = False
    drop = False
    jumpable = False
    grabbable = False
    isObstacle = False

    obstacleId = None

    def getAngle(self):
        '''
        Must return the angle of this obstacle. Used when a shot bounces off
        this obstacle.
        '''

    def collide(self, unit, deltaX, deltaY):
        '''
        Called to see whether the given unit would collide with this obstacle
        given its trajectory. Returns the original (deltaX, deltaY) if the unit
        would not collide with the obstacle, or if the unit would collide with
        the obstacle, returns the new (deltaX, deltaY) to bring the unit into
        contact with the obstacle.
        '''

    def finalPosition(self, unit, deltaX, deltaY):
        '''
        Returns the final position of the given unit which is attempting to
        travel along the given displacement.  May assume that the object hits
        this obstacle on the way.  If the final position is simply the
        collision point, should return None.
        '''
        return None

    def hitByPlayer(self, player):
        pass

    def walkTrajectory(self, vel, deltaTime):
        '''
        Returns the displacement that a player would be going on this
        surface if they would travel, given an absolute velocity and an amount
        of time.

        Only relevant if ground is True.
        '''
        raise NotImplementedError(
            '%s.walkTrajectory' % (self.__class__.__name__,))

    def checkBounds(self, unit, deltaX, deltaY):
        '''
        Checks whether the given point is beyond the bounds of this piece of
        ground. Returns (obstacle, pos) where obstacle is this or another
        ground obstacle that the player has walked to, or None if the player
        has walked off the ground altogether.

        Only relevant if ground is True.
        '''

    def getPosition(self):
        '''
        Should return a point approximately at the centre of this obstacle.
        '''

    def discretise(self, unit, pos):
        '''
        Used for AI path-finding. Should return a position that's approximately
        equal to the given position. Only relevant if jumpable is True.
        '''
        return self.getPositionFromIndex(
            unit, self.getPositionIndex(unit, pos))

    def getPositionIndex(self, unit, pos):
        '''
        Also used for AI path-finding. Should return an integer indicating
        how far along this obstacle the given position lies. Only relevant if
        jumpable is True.
        '''
        raise NotImplementedError(
            '%s.getPositionIndex' % (self.__class__.__name__),)

    def getPositionFromIndex(self, unit, posIndex):
        '''
        Should convert from the given position index to the corresponding
        discretised position. Only relevant if jumpable is True.
        '''
        raise NotImplementedError(
            '%s.getPositionFromIndex' % (self.__class__.__name__),)

    def getPositionIndexBounds(self, unit):
        '''
        Only relevant if jumpable is True. Returns (minIndex, maxIndex) showing
        the bounds on valid position indices for this obstacle.
        '''
        raise NotImplementedError(
            '%s.getPositionIndexBounds' % (self.__class__.__name__),)

    def setId(self, obstacleId):
        '''
        Gives this obstacle an index that is unique within the map block
        layout.
        '''
        self.obstacleId = obstacleId


class Obstacle(Obstaclish):
    '''Represents an obstacle which can't be passed by player or shot.'''
    isObstacle = True

    def __init__(self, pt1, deltaPt):
        self._angle = None
        self.pt1 = pt1
        self.deltaPt = deltaPt

        if deltaPt[0] > 0:
            y = +1
        elif deltaPt[0] < 0:
            y = -1
        else:
            y = 0
        if deltaPt[1] > 0:
            x = -1
        elif deltaPt[1] < 0:
            x = +1
        else:
            x = 0
        self._unitCornerOffset = (x, y)

    @property
    def pt2(self):
        return (self.pt1[0] + self.deltaPt[0],
                self.pt1[1] + self.deltaPt[1])

    def getPosition(self):
        return (self.pt1[0] + 0.5 * self.deltaPt[0],
                self.pt1[1] + 0.5 * self.deltaPt[1])

    def getAngle(self):
        if self._angle is None:
            self._angle = atan2((self.deltaPt[1] + 0.), (self.deltaPt[0] + 0.))
        return self._angle

    def collide(self, unit, deltaX, deltaY, soft=False):
        '''
        Called to see whether the given unit would collide with this obstacle
        given its trajectory. Returns the original (deltaX, deltaY) if the unit
        would not collide with the obstacle, or if the unit would collide with
        the obstacle, returns the new (deltaX, deltaY) to bring the unit into
        contact with the obstacle.
        '''
        unitCorner = (
            unit.pos[0] + self._unitCornerOffset[0] * unit.HALF_WIDTH,
            unit.pos[1] + self._unitCornerOffset[1] * unit.HALF_HEIGHT)
        result = collideTrajectory(
            self.pt1, self.deltaPt, unitCorner, deltaX, deltaY,
            soft=soft, radius=unit.HALF_WIDTH)
        if result is None:
            return deltaX, deltaY
        return result


def joinObstacles(obs1, obs2):
    '''
    Takes two Obstacle objects which are to be joined in a clockwise direction.
    Sets whatever attributes are required to join them, and returns a (possibly
    empty) collection of Corner objects which need to be added in order for
    this join to occur.
    '''
    def getQuadrant(obs):
        dX, dY = obs.deltaPt
        if dY > 0:
            if dX > 0:
                return 1
            elif dX == 0:
                return 1.5
            else:
                return 2
        elif dY < 0:
            if dX < 0:
                return 3
            elif dX == 0:
                return 3.5
            else:
                return 4
        else:
            if dX < 0:
                return 2.5
            else:
                return 0.5

    q1 = getQuadrant(obs1)
    q2 = getQuadrant(obs2)
    pt = obs2.pt1

    if q1 == q2 and (q1 == 4 or q1 == 1):
        obs1.rightGround = obs2
        obs2.leftGround = obs1
        return []

    d = (q1 - q2) % 4
    if d < 2:
        # Concave.
        return []

    if q1 == q2:
        if ((obs1.getAngle() - obs2.getAngle()) % (2 * pi)) < pi:
            # Also concave.
            return []

    result = []
    if q1 in (2, 2.5, 3, 3.5) and q2 in (3.5, 4, 0.5, 1):
        result.append(Corner(pt, (1, -1), (0, 2)))
    if q1 in (3, 3.5, 4, 0.5) and q2 in (0.5, 1, 1.5, 2):
        c = CornerTop(pt)
        result.append(c)
        if obs1.deltaPt[0] > 0:
            obs1.rightGround = c
            c.leftGround = obs1
            assert isinstance(obs1, GroundObstacle)
        if obs2.deltaPt[0] > 0:
            c.rightGround = obs2
            obs2.leftGround = c
            assert isinstance(obs2, GroundObstacle)
    if q1 in (4, 0.5, 1, 1.5) and q2 in (1.5, 2, 2.5, 3):
        result.append(Corner(pt, (-1, 1), (0, -2)))
    if q1 in (1, 1.5, 2, 2.5) and q2 in (2.5, 3, 3.5, 4):
        result.append(Corner(pt, (-1, -1), (2, 0)))
    return result


class Corner(Obstaclish):
    '''
    Note that offset and delta are defined for Corners in the reverse of the
    way in which they are defined for obstacles. This is because Corner
    collision detection collides the corner with the unit rather than the other
    way around.
    '''

    def __init__(self, pt, offset, delta):
        self.pt = pt
        self.offset = offset
        self.delta = delta
        self._angle = atan2((-offset[1] + 0.), (-offset[0] + 0.))

    def getPosition(self):
        return self.pt

    def getAngle(self):
        '''
        Must return the angle of this obstacle. Used when a shot bounces off
        this obstacle.
        '''
        return self._angle

    def collide(self, unit, deltaX, deltaY):
        if self.delta == (0, 2):
            if deltaX <= 0:
                return deltaX, deltaY
        elif self.delta == (0, -2):
            if deltaX >= 0:
                return deltaX, deltaY
        elif self.delta == (2, 0):
            if deltaY >= 0:
                return deltaX, deltaY
        elif self.delta == (-2, 0):
            if deltaY <= 0:
                return deltaX, deltaY

        oPt1 = (unit.pos[0] + self.offset[0] * unit.HALF_WIDTH,
                unit.pos[1] + self.offset[1] * unit.HALF_HEIGHT)
        oDeltaPt = (
            self.delta[0] * unit.HALF_WIDTH,
            self.delta[1] * unit.HALF_HEIGHT)

        result = collideTrajectory(
            oPt1, oDeltaPt, self.pt, -deltaX, -deltaY,
            radius=2, soft=True)
        if result is not None:
            deltaX, deltaY = result
            return -deltaX, -deltaY
        return deltaX, deltaY

    def hitByPlayer(self, player):
        # Stop any upward motion of the player.
        player.yVel = max(player.yVel, 0)
        player._jumpTime = 0

    def finalPosition(self, unit, deltaX, deltaY):
        if self.delta[0] == 0:
            return unit.pos[0], unit.pos[1] + (deltaY / 2)

        if deltaY == 0:
            return None
        return unit.pos[0], unit.pos[1] + deltaY


def playerHitJumpable(self, player):
    '''This obstacle has been hit by player.'''
    if player.world.physics.playerBounce:
        player._jumpTime = player.world.physics.playerMaxJumpTime
        player.yVel = -player.yVel
        player.detachFromEverything()
    else:
        player._jumpTime = 0
        player.yVel = 0


class CornerTop(Corner):
    ground = True
    jumpable = True

    def __init__(self, pt):
        Corner.__init__(self, pt, (1, 1), (-2, 0))
        self.rightGround = None
        self.leftGround = None
        self.subshape = ()

    def checkBounds(self, unit, deltaX, deltaY):
        '''
        Only relevant for the ground section.
        '''
        pt1x = self.pt[0] - unit.HALF_WIDTH
        pt1y = self.pt[1] - unit.HALF_HEIGHT
        pt2x = self.pt[0] + unit.HALF_WIDTH
        pt2y = pt1y

        posX = unit.pos[0] + deltaX
        posY = unit.pos[1] + deltaY
        return checkGroundBounds(
            self, (pt1x, pt1y), (pt2x, pt2y), (posX, posY))

    def walkTrajectory(self, vel, deltaTime):
        return (vel * deltaTime, 0)

    hitByPlayer = playerHitJumpable

    def finalPosition(self, unit, deltaX, deltaY):
        return None

    def getPositionIndex(self, unit, pos):
        return int(floor((pos[0] - self.pt[0]) / D_UNIT + 0.5))

    def getPositionFromIndex(self, unit, posIndex):
        x = self.pt[0] + posIndex * D_UNIT
        minX = self.pt[0] - unit.HALF_WIDTH + 0.1
        maxX = self.pt[0] + unit.HALF_HEIGHT - 0.1
        x = min(maxX, max(minX, x))
        y = self.pt[1] - unit.HALF_HEIGHT
        return (x, y)

    def getPositionIndexBounds(self, unit):
        y = 0   # Doesn't matter because of how discretise() works
        minI = self.getPositionIndex(unit, (self.pt[0] - unit.HALF_HEIGHT, y))
        maxI = self.getPositionIndex(unit, (self.pt[0] + unit.HALF_HEIGHT, y))
        return (minI, maxI)


class JumpableObstacle(Obstacle):
    '''Represents an obstacle that is not a wall.'''
    jumpable = True
    hitByPlayer = playerHitJumpable


class GroundObstacle(JumpableObstacle):
    '''Represents an obstacle that players are allowed to walk on.'''
    ground = True

    def __init__(self, pt1, deltaPt):
        super(GroundObstacle, self).__init__(pt1, deltaPt)
        angle = self.getAngle()
        self.ratio = (cos(angle), sin(angle))
        self.leftGround = None
        self.rightGround = None
        self.subshape = ()

    def walkTrajectory(self, vel, deltaTime):
        '''Returns the displacement that a player would be going on this
        surface if they would travel, given an absolute velocity and an amount
        of time.'''
        return tuple([vel * deltaTime * self.ratio[i] for i in (0, 1)])

    def checkBounds(self, unit, deltaX, deltaY):
        pt1x = self.pt1[0] - self._unitCornerOffset[0] * unit.HALF_WIDTH
        pt1y = self.pt1[1] - self._unitCornerOffset[1] * unit.HALF_HEIGHT
        pt2x = pt1x + self.deltaPt[0]
        pt2y = pt1y + self.deltaPt[1]

        posX = unit.pos[0] + deltaX
        posY = unit.pos[1] + deltaY

        return checkGroundBounds(
            self, (pt1x, pt1y), (pt2x, pt2y), (posX, posY))

    def finalPosition(self, unit, deltaX, deltaY):
        '''Returns the final position of the given unit
        trying to travel a displacement of [deltaX, deltaY], given
        that it collides with this obstacle.
        This routine takes into account that if the object is travelling
        upwards, it may slide up the obstacle. Returns None if no sliding is
        going on, to indicate that the object hits the ground.'''

        if deltaY > -0.001 or self.deltaPt[1] == 0:
            return None

        # We collided with the floor while jumping.
        pt1x = self.pt1[0] - self._unitCornerOffset[0] * unit.HALF_WIDTH
        pt1y = self.pt1[1] - self._unitCornerOffset[1] * unit.HALF_HEIGHT

        # Go to the y position we would have gone to.
        y = unit.pos[1] + deltaY

        # Calculate where this lies on the slope.
        x = pt1x + (y - pt1y) * (
            self.deltaPt[0] / (self.deltaPt[1] + 0.))

        return x, y

    def getPositionIndex(self, unit, pos):
        # TODO: maybe calculate from the middle, not the corner
        pt1x = self.pt1[0] - self._unitCornerOffset[0] * unit.HALF_WIDTH
        pt1y = self.pt1[1] - self._unitCornerOffset[1] * unit.HALF_HEIGHT

        dist = math.distance((pt1x, pt1y), pos)
        fullDist = math.distance((0, 0), self.deltaPt)
        midPoint = 0.5 * fullDist
        return int(floor((dist - midPoint) / D_UNIT + 0.5))

    def getPositionFromIndex(self, unit, posIndex):
        pt1x = self.pt1[0] - self._unitCornerOffset[0] * unit.HALF_WIDTH
        pt1y = self.pt1[1] - self._unitCornerOffset[1] * unit.HALF_HEIGHT
        fullDist = math.distance((0, 0), self.deltaPt)
        midPoint = 0.5 * fullDist

        u = midPoint + posIndex * D_UNIT
        f = min(0.999, max(0.001, u / fullDist))
        x = pt1x + f * self.deltaPt[0]
        y = pt1y + f * self.deltaPt[1]
        return (x, y)

    def getPositionIndexBounds(self, unit):
        minX = self.pt1[0] - self._unitCornerOffset[0] * unit.HALF_WIDTH
        minY = self.pt1[1] - self._unitCornerOffset[1] * unit.HALF_HEIGHT
        maxX = minX + self.deltaPt[0]
        maxY = minY + self.deltaPt[1]
        minI = self.getPositionIndex(unit, (minX, minY))
        maxI = self.getPositionIndex(unit, (maxX, maxY))
        return (minI, maxI)


class LedgeObstacle(GroundObstacle):
    drop = True

    def collide(self, unit, deltaX, deltaY):
        unitCorner = (
            unit.pos[0] + self._unitCornerOffset[0] * unit.HALF_WIDTH,
            unit.pos[1] + self._unitCornerOffset[1] * unit.HALF_HEIGHT)
        result = collideTrajectory(
            self.pt1, self.deltaPt, unitCorner, deltaX, deltaY)
        if result is None:
            return deltaX, deltaY
        return result


class RoofObstacle(Obstacle):
    def __init__(self, pt1, deltaPt):
        super(RoofObstacle, self).__init__(pt1, deltaPt)

    def finalPosition(self, unit, deltaX, deltaY):
        '''Returns the final position of a point-sized object at the specified
        position trying to travel a displacement of [deltaX, deltaY], given
        that it collides with this obstacle.'''
        # If an object collides with a roof obstacle while falling, it still
        #  falls.

        if deltaY > 0 and deltaX != 0 and self.deltaPt[1] != 0:
            # We collided with the roof while falling.
            # New position is where the roof is at the correct y-position.

            pt1x = self.pt1[0] - self._unitCornerOffset[0] * unit.HALF_WIDTH
            pt1y = self.pt1[1] - self._unitCornerOffset[1] * unit.HALF_HEIGHT

            x = (pt1x + (unit.pos[1] + deltaY - pt1y) *
                 self.deltaPt[0] / (0. + self.deltaPt[1]) -
                 ((deltaX / abs(deltaX)) * 0.1))
            y = unit.pos[1] + deltaY
            # Note: the above statement should only ever cause a division by
            #  zero if someone has been silly enough to put a piece of
            #  horizontal roof in upside down (impassable from above not below)
            return x, y

        # Normal case:
        return None

    def hitByPlayer(self, player):
        # Stop any upward motion of the player.
        player.yVel = max(player.yVel, 0)
        player._jumpTime = 0


class FillerRoofObstacle(RoofObstacle):
    '''Represents an obstacle that would be used as an obstacleEdge;
    its endpoint calculations are less precise (that being they
    don't exist), but are able to be so due to their usage.'''

    def finalPosition(self, unit, deltaX, deltaY):
        if self.deltaPt[0] == 0:
            return unit.pos[0], unit.pos[1] + (deltaY / 2)
        else:
            super(FillerRoofObstacle, self).finalPosition(unit, deltaX, deltaY)

    def collide(self, unit, deltaX, deltaY, soft=True):
        '''Returns how far a point-sized object at the specified position
        could travel if it were trying to travel a displacement of
        [deltaX, deltaY].
        '''
        # Note: This routine is deliberately slightly different to the default
        # collide() in that it has less tolerance for rounding error around
        # corners.
        return super(FillerRoofObstacle, self).collide(
            unit, deltaX, deltaY, soft=True)


class VerticalWall(JumpableObstacle):
    '''Represents a vertical wall that players can cling to and climb.'''
    drop = True
    grabbable = True

    def __init__(self, pt1, deltaPt):
        assert deltaPt[0] == 0

        # Can drop off a vertical wall, so drop is always set to True
        super(VerticalWall, self).__init__(pt1, deltaPt)

    def unstickyWallFinalPosition(self, pt, deltaX, deltaY):
        '''Returns the final position of a point-sized object at the specified
        position trying to travel a displacement of [deltaX, deltaY], given
        that it collides with this obstacle.'''
        # If an object collides with a roof obstacle while falling, it still
        #  falls.

        if deltaY == 0:
            return None

        return pt[0], pt[1] + deltaY

    def getPositionIndex(self, unit, pos):
        x = self.pt1[0]
        if self.deltaPt[1] > 0:
            x += unit.HALF_WIDTH
        else:
            x -= unit.HALF_WIDTH

        midPoint = self.pt1[1] + 0.5 * self.deltaPt[1]
        return int(floor((pos[1] - midPoint) / D_UNIT + 0.5))

    def getPositionFromIndex(self, unit, posIndex):
        x = self.pt1[0]
        if self.deltaPt[1] > 0:
            x += unit.HALF_WIDTH
        else:
            x -= unit.HALF_WIDTH

        y2 = self.pt1[1] + self.deltaPt[1]
        minY = min(self.pt1[1], y2) + 0.1
        maxY = max(self.pt1[1], y2) - 0.1

        midPoint = self.pt1[1] + 0.5 * self.deltaPt[1]
        y = midPoint + posIndex * D_UNIT
        y = min(maxY, max(minY, y))

        return (x, y)

    def getPositionIndexBounds(self, unit):
        x = 0   # Doesn't matter because of how discretise() works
        y2 = self.pt1[1] + self.deltaPt[1]
        minY = min(self.pt1[1], y2)
        maxY = max(self.pt1[1], y2)
        minI = self.getPositionIndex(unit, (x, minY))
        maxI = self.getPositionIndex(unit, (x, maxY))
        return (minI, maxI)


def collideTrajectory(
        obstaclePt1, obstacleDeltaPt, pt, deltaX, deltaY,
        soft=False, radius=0):
    '''
    Calculates whether the trajectory specified by pt, deltaX and deltaY would
    hit the given obstacle.

    @param soft: if true, there's less tolerance for rounding errors around
            corners, making it harder to hit the corners of the obstacle.
    @param radius: this parameter is used to simulate the moving point having
            actual size. If the give point is just on the "solid" side of the
            given obstacle, then providing a radius will ensure that the point
            does not continue to fall through that obstacle.
    @returns: None if no collision would occur, (deltaX, deltaY) otherwise.
    '''

    if soft:
        epsilon = 0
    else:
        epsilon = 1e-10

    ax, ay = obstaclePt1
    bx, by = pt
    dX1, dY1 = obstacleDeltaPt

    # Check if the lines are parallel.
    denom = dX1 * deltaY - deltaX * dY1
    if denom <= epsilon:
        # We can go through it in this direction.
        return None

    # Calculate whether the line segments intersect.
    if deltaX == 0 and deltaY == 0:
        return None
    elif radius > 0:
        minT = radius / (deltaX ** 2 + deltaY ** 2) ** 0.5
    else:
        minT = 0

    t = (dX1 * (ay - by) - dY1 * (ax - bx)) / denom
    if t < (-minT - 1e-10) or t > (1. + 1e-10):
        return None

    if t < -1e10:
        epsilon = 0

    s = (deltaX * (ay - by) - deltaY * (ax - bx)) / denom
    # Take into account floating point error.
    if s <= -epsilon or s >= (1. + epsilon):
        # Past the end of the obstacle.
        return None

    # Calculate the actual collision point.
    x = bx + t * deltaX
    y = by + t * deltaY

    # Calculate the allowed change in position.
    resultX, resultY = x - pt[0], y - pt[1]

    # Ensure that the tollerance doesn't make us go in the opposite direction
    # to the way we're trying to move.
    if resultX * deltaX < 0 or resultY * deltaY < 0:
        return 0, 0
    return resultX, resultY


def checkGroundBounds(obstaclish, pt1, pt2, unitPos):
    pt1x, pt1y = pt1
    pt2x, pt2y = pt2

    posX, posY = unitPos

    if posX < pt1x:
        if obstaclish.leftGround is None:
            return None, (posX, posY)
        return obstaclish.leftGround, (pt1x, pt1y)
    elif posX > pt2x:
        if obstaclish.rightGround is None:
            return None, (posX, posY)
        return obstaclish.rightGround, (pt2x, pt2y)
    return obstaclish, (posX, posY)



import imp
import os.path
import sys
from setuptools import setup, find_packages

additional_dlls = [
    'pygame/SDL_ttf.dll',
    'pygame/libvorbis-0.dll',
    'pygame/libvorbisfile-3.dll',
    'pygame/libfreetype-6.dll',
    'pygame/SDL_mixer.dll',
    'pygame/libogg-0.dll',
]


here = os.path.abspath(os.path.dirname(__file__))
trosnoth_version = imp.load_source('trosnoth.version', os.path.join(
    here, 'trosnoth', 'version.py')).version


def main():
    if 'py2exe' in sys.argv:
        import py2exe

        # Make sure py2exe knows which data files to include.
        paths = [
	    'trosnoth/data',
            'trosnoth/data/achievements',
            'trosnoth/data/blocks',
            'trosnoth/data/config',
            'trosnoth/data/fonts',
            'trosnoth/data/music',
            'trosnoth/data/sound',
            'trosnoth/data/sprites',
            'trosnoth/data/startupMenu',
            'trosnoth/data/statGeneration',
            'trosnoth/data/themes',
            'trosnoth/data/themes/pirate',
            'trosnoth/data/themes/pirate/blocks',
            'trosnoth/data/themes/pirate/config',
            'trosnoth/data/themes/pirate/fonts',
            'trosnoth/data/themes/pirate/sprites',
            'trosnoth/data/themes/pirate/startupMenu',
            'trosnoth/data/web',
        ]

        data = []
        for path in paths:
            files = []
            for filename in os.listdir(path):
                if filename in ('__init__.py', '__init__.pyc'):
                    continue
                fn = os.path.join(path, filename)
                if os.path.isfile(fn):
                    files.append(fn)
            data.append((path, files))

        moreargs = {
            'console': [
                {'script': 'scripts/trosnoth',
                    'icon_resources': [(1, 'wininstall/icon.ico')]
                },
                'scripts/trosnoth-server',
            ],
            'data_files': data,
            'options': {
                'py2exe': {
                    'includes': 'zope.interface,pygame._view,trosnoth.ais.alpha,trosnoth.ais.simple,trosnoth.ais.john,trosnoth.ais.test,trosnoth.ais.ranger',
                },
            },
       }
    else:
        moreargs = {}

    setup(name = 'trosnoth',
        version = trosnoth_version,
        description = 'Trosnoth network platform game',
        author = 'J.D. Bartlett et al',
        author_email = 'josh@trosnoth.org',
        url = 'http://www.trosnoth.org/',
        packages=find_packages(exclude=['test']),

        # Mapping says which files each package needs.
        package_data = {
            'trosnoth.data.blocks': ['*.block', '*.png', '*.bmp'],
            'trosnoth.data.fonts': ['*.ttf', '*.TTF', '*.txt'],
            'trosnoth.data.music': ['*.ogg'],
            'trosnoth.data.sound': ['*.ogg'],
            'trosnoth.data.sprites': ['*.png', '*.bmp'],
            'trosnoth.data.startupMenu': ['*.png', '*.txt'],
            'trosnoth.data.statGeneration': ['*.htm'],
            'trosnoth.data.themes': ['pirate/info.txt',
                'pirate/blocks/*.png', 'pirate/config/*.cfg',
                'pirate/fonts/*', 'pirate/sprites/*',
                'pirate/startupMenu/*'],
            'trosnoth.data': [
                'config/*.cfg', 'achievements/*.png', 'web/*.png',
                'pathfinding.db'],
            'trosnoth': ['gpl.txt']
        },

        scripts = ['scripts/trosnoth', 'scripts/trosnoth-server'],
        long_description = 'Trosnoth is a very very addictive and fun network team game.' ,

        install_requires = [
            'pygame',
            'twisted>=10.0',
        ],

        classifiers = [
            'Development Status :: 5 - Production/Stable',
            'Environment :: MacOS X',
            'Environment :: Win32 (MS Windows)',
            'Environment :: X11 Applications',
            'Framework :: Twisted',
            'Intended Audience :: End Users/Desktop',
            'License :: OSI Approved :: GNU General Public License (GPL)',
            'Natural Language :: English',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Topic :: Games/Entertainment :: Arcade',
            'Topic :: Games/Entertainment :: Side-Scrolling/Arcade Games',
        ],
        **moreargs
    ) 


if __name__ == '__main__':
    main()



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from trosnoth.gui.framework.framework import CompoundElement
from trosnoth.gui.common import Relative

class Tab(CompoundElement):
    borderWidth = 4
    def __init__(self, app, caption):
        super(Tab, self).__init__(app)
        self.caption = caption
        self.container = None

    def Relative(self, x, y):
        '''
        Returns an object representing a position relative to this tab's canvas.
        See common.Relative() for more information.
        '''
        def getRect():
            return self.container._getTabInternalRect()
        return Relative(getRect, x, y)

    def activated(self):
        pass

    def beforeDeactivating(self):
        return True

    def deactivated(self):
        pass

    def __str__(self):
        return self.caption

    def __repr__(self):
        return self.caption



'''universe.py - defines anything that has to do with the running of the
universe. This includes players, shots, zones, and the level itself.'''

from collections import defaultdict
import functools
import heapq
import logging
from math import sin, cos
import random
import itertools

from twisted.internet import defer, reactor

from trosnoth.const import (
    TURRET_DEATH, TROSBALL_DEATH, SHOXWAVE_DEATH, STALEMATE_CHECK_PERIOD,
    COLLECTABLE_STAR_LIFETIME, NO_COLLECTABLE_STARS, TICK_PERIOD,
    MAX_PLAYER_NAME_LENGTH, ZONE_TAG_STAR_HALF_LIFE, GAME_STATE_CHECK_PERIOD,
    RABBIT_CHASING_PERIOD, BOMBER_DEATH,
)
from trosnoth.utils import globaldebug
from trosnoth.utils.math import distance
from trosnoth.model.idmanager import IdManager
from trosnoth.model.upgrades import upgradeOfType, allUpgrades
from trosnoth.model.physics import WorldPhysics

from trosnoth.model import gameStates
from trosnoth.model.map import MapLayout, MapState
from trosnoth.model.player import Player
from trosnoth.model.star import CollectableStar
from trosnoth.model.shot import Shot, GrenadeShot
from trosnoth.model.team import Team
from trosnoth.model.trosball import Trosball

from trosnoth.utils.event import Event
from trosnoth.utils.message import MessageConsumer
from trosnoth.utils.twist import WeakCallLater, WeakLoopingCall
from trosnoth.utils.unrepr import unrepr
from trosnoth.messages import (
    TaggingZoneMsg, ShotFiredMsg, RespawnMsg, ChatFromServerMsg,
    GameStartMsg, GameOverMsg, RemovePlayerMsg, ShotHitPlayerMsg,
    AddPlayerMsg, PlayerStarsSpentMsg, SetPlayerTeamMsg, WorldLoadingMsg,
    ZoneStateMsg, WorldResetMsg,
    CreateCollectableStarMsg, RemoveCollectableStarMsg, PlayerHasElephantMsg,
    FireShoxwaveMsg, UpgradeChangedMsg,
    PlayerHasTrosballMsg, TrosballPositionMsg, AwardPlayerStarMsg,
    TrosballTagMsg, ChangeTimeLimitMsg, StartingSoonMsg, TickMsg,
    ReturnToLobbyMsg, TICK_LIMIT,
)

log = logging.getLogger('universe')


class DelayedCall(object):
    def __init__(self, fn, args, kwargs):
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.cancelled = False

    def cancel(self):
        self.cancelled = True



class UnknownMapLayouts(Exception):
    def __init__(self, *args, **kwargs):
        super(UnknownMapLayouts, self).__init__(*args, **kwargs)
        self.keys = self.message


class Timer(object):
    def __init__(self, world):
        self.world = world
        self.duration = 0
        self.gameStartingIn = -1
        self.gameStartDelay = -1
        self.gameOverTime = None
        self.elapsedTime = 0.0

        self.onGameStart = Event()

    def setMatchDuration(self, duration):
        self.duration = duration

    def getMatchDuration(self):
        return self.duration

    def hasTimeLimit(self):
        if not self.world.state.isTimedGameState():
            return False
        return self.duration != 0

    def getTimeLeft(self):
        '''
        Return the amount of time left on the clock. If the game is over, the
        clock may be stopped at the figure it was on when the game finished.
        If the game is untimed, may return 0.
        '''
        gameTimeMode = self.world.state.getGameTimeMode()
        if gameTimeMode == 'unstarted':
            return self.duration
        if self.duration == 0:
            # Untimed game
            return 0
        if gameTimeMode == 'running':
            return self.duration - self.getElapsedTime()
        if gameTimeMode == 'ended':
            return self.duration - self.gameOverTime

    def dumpState(self):
        return {
            'timeRunning': self.getElapsedTime(),
            'timeLimit': self.duration,
        }

    def restoreState(self, data):
        self.duration = data['timeLimit']
        self.elapsedTime = data['timeRunning']

    def tick(self):
        if self.gameStartingIn > 0:
            self.gameStartingIn -= TICK_PERIOD
            if self.gameStartingIn <= 0:
                self.onGameStart()
                self.onGameStart = Event()

        self.elapsedTime += self.world.tickPeriod

    def beginGameStartCountdown(self, delay):
        self.gameStartingIn = delay
        self.gameStartDelay = delay

    def getStartCountdown(self):
        return self.gameStartingIn

    def getStartCountdownMax(self):
        return self.gameStartDelay

    def gameFinishedNow(self):
        self.gameOverTime = self.getElapsedTime()

    def getElapsedTime(self):
        return self.elapsedTime

    def reset(self):
        self.elapsedTime = 0.0


class Abilities(object):
    def __init__(self):
        self.upgrades = True
        self.respawn = True
        self.leaveFriendlyZones = True
        self.zoneCaps = True
        self.stars = True


class UIOptions(object):
    '''
    Represents options related to how this kind of match should be displayed to
    the user.
    '''

    def __init__(self):
        self.showNets = False
        self.shiftingFrontLine = False
        self.frontLineGetter = None
        self.showWinner = False
        self.showReadyStates = False


class StarSpawner(object):
    def __init__(self, world):
        self.world = world

        self._lastStalemateCheck = 0.0
        self.lastZoneTagged = 0.0

        self.world.onZoneTagged.addListener(self.zoneTagged)

    def stop(self):
        self.world.onZoneTagged.removeListener(self.zoneTagged)

    def reset(self):
        self._lastStalemateCheck = 0.0
        self.lastZoneTagged = 0.0

    def zoneTagged(self, *args, **kwargs):
        self.lastZoneTagged = self.world.timer.getElapsedTime()

    def collectableStarsCanAppear(self):
        if NO_COLLECTABLE_STARS:
            return False
        if not self.world.abilities.stars:
            return False
        return True

    def tick(self):
        if not self.collectableStarsCanAppear():
            return

        now = self.world.timer.getElapsedTime()
        if now < self._lastStalemateCheck + STALEMATE_CHECK_PERIOD:
            return
        self._lastStalemateCheck = now

        prob = 0.5 ** ((now - self.lastZoneTagged) / ZONE_TAG_STAR_HALF_LIFE)
        if random.random() > prob:
            # Distribute the stars evenly by team
            team = random.choice(self.world.teams)
            zone = self.world.selectZoneForTeam(team.id)
            if zone.owner != team:
                return
            x, y = zone.defn.randomPosition()
            starId = self.world.idManager.newStarId()
            if starId is not None:
                self.world.sendServerCommand(
                    CreateCollectableStarMsg(starId, x, y))



class Universe(MessageConsumer):
    '''Universe(halfMapWidth, mapHeight)
    Keeps track of where everything is in the level, including the locations
    and states of every alien, the terrain positions, and who owns the
    various territories and orbs.'''

    PLAYER_RESET_STARS = 0

    isServer = False

    def __init__(
            self, layoutDatabase, halfMapWidth=3, mapHeight=2,
            authTagManager=None, duration=0, onceOnly=False):
        '''
        halfMapWidth:   is the number of columns of zones in each team's
                        territory at the start of the game. There will always
                        be a single column of neutral zones between the two
                        territories at the start of the game.
        mapHeight:      is the number of zones in every second column of
                        zones. Every other column will have mapHeight + 1
                        zones in it. This is subject to the constraints that
                        (a) the columns at the extreme ends of the map will
                        have mapHeight zones; (b) the central (initially
                        neutral) column of zones will never have fewer zones
                        in it than the two bordering it - this will sometimes
                        mean that the column has mapHeight + 2 zones in it.
        '''
        super(Universe, self).__init__()

        self.onPlayerAdded = Event()        # (player)
        self.onPlayerRemoved = Event()      # (player, oldId)
        self.onGameStateChanged = Event()
        self.onTeamScoreChanged = Event()
        self.onShotRemoved = Event()        # (shotId)
        self.onCollectableStarRemoved = Event()
        self.onSectorNeutralised = Event()  # (tagger, zoneCount)
        self.onGameOver = Event()           # (team, timeOver)
        self.onZoneTagged = Event()         # (zone, player, previousOwner)
        self.onOpenChatReceived = Event()   # (text, sender)
        self.onTeamChatReceived = Event()   # (team, text, sender)
        self.onPlayerKill = Event()         # (killer, target, deathType)
        self.onPlayerRespawn = Event()      # (player)
        self.onGrenadeExplosion = Event()   # (pos, radius)
        self.onTrosballExplosion = Event()  # (player)
        self.onBomberExplosion = Event()    # (player)
        self.onReset = Event()
        self.onServerTickComplete = Event()
        self.onChangeVoiceChatRooms = Event()   # (teams, players)

        self.delayedCalls = []

        self.isIncomplete = False
        self.playerWithElephant = None
        self.physics = WorldPhysics(self)
        if authTagManager is None:
            self.authManager = None
        else:
            self.authManager = authTagManager.authManager

        self.playerWithId = {}
        self.shotWithId = {}
        self.teamWithId = {'\x00': None}

        self.trosballScore = {'A': 0, 'B': 0}

        # Create Teams:
        self.teams = (
            Team(self, 'A'),
            Team(self, 'B'),
        )
        Team.setOpposition(self.teams[0], self.teams[1])

        for t in self.teams:
            self.teamWithId[t.id] = t

        # Set up zones
        self.zoneWithDef = {}
        self.layout = MapLayout(halfMapWidth, mapHeight)
        self.map = MapState(self, self.layout)
        self.zoneWithId = self.map.zoneWithId
        self.zones = self.map.zones
        self.zoneBlocks = self.map.zoneBlocks

        self.players = set()
        self.grenades = set()
        self.collectableStars = {}      # starId -> CollectableStar
        self.trosball = None
        self.deadStars = set()
        self.trosballPlayer = None
        self.trosballCooldownPlayer = None
        self.playerGotTrosballTime = None
        self.gameMode = 'Normal'
        self.rogueTeamName = 'Rogue'
        self.tickPeriod = TICK_PERIOD
        self._gameSpeed = 1.0
        self._gameInProgress = False
        self._winningTeam = None
        self.lastTickId = 0
        self.loading = False

        self.layoutDatabase = layoutDatabase
        self._onceOnly = onceOnly
        self.timer = Timer(self)
        self.timer.setMatchDuration(duration)
        self.abilities = Abilities()
        self.uiOptions = UIOptions()

        self.loadedMap = None

        self.setGameState(gameStates.Lobby)
        self._nextGameStateCheck = self.timer.getElapsedTime()

    def isValidNick(self, nick):
        if len(nick) < 2 or len(nick) > MAX_PLAYER_NAME_LENGTH:
            return False
        return True

    @property
    def shots(self):
        '''
        Used by UI to iterate through shots.
        '''
        return self.shotWithId.itervalues()

    def defaultHandler(self, msg):
        msg.applyOrderToWorld(self)

    def selectZoneForTeam(self, teamId):
        '''
        Randomly selects a zone, giving preference to:
            1. Zones owned by the given team that are adjacent to an enemy zone
            2. Zones owned by the given team that are adjacent to a zone not
                owned by the team.
            3. Other zones owned by the given team.
            4. Other zones.
        '''
        team = self.getTeam(teamId)
        allTeamZones = [
            z for z in self.map.zones
            if z.owner is not None and z.owner.id == teamId]
        nextToEnemy = []
        nextToNeutral = []
        for zone in list(allTeamZones):
            enemy = neutral = False
            for adjDef in zone.defn.adjacentZones.iterkeys():
                if adjDef is None:
                    continue
                adj = self.zoneWithDef[adjDef]
                if adj.owner is None:
                    neutral = True
                elif adj.isEnemyTeam(team):
                    enemy = True
            if enemy:
                nextToEnemy.append(zone)
            elif neutral:
                nextToNeutral.append(zone)

        return random.choice(
            nextToEnemy
            or nextToNeutral
            or allTeamZones
            or list(self.map.zones))

    @WorldResetMsg.handler
    def gotWorldReset(self, msg):
        if not self.isServer:
            data = unrepr(msg.settings)
            self.restoreEverything(data)

        self.timer.reset()
        self._nextGameStateCheck = GAME_STATE_CHECK_PERIOD

    def getTeam(self, teamId):
        if teamId == '\x00':
            return None
        return self.teamWithId[teamId]

    def getPlayer(self, playerId, default=None):
        return self.playerWithId.get(playerId, default)

    def getUpgradeType(self, upgradeTypeId):
        return upgradeOfType[upgradeTypeId]

    def getZone(self, zoneId, default=None):
        return self.map.zoneWithId.get(zoneId, default)

    def getShot(self, sId):
        return self.shotWithId[sId]

    def setGameMode(self, mode):
        if self.physics.setMode(mode):
            self.gameMode = mode
            log.debug('Client: GameMode is set to ' + mode)

    def setGameSpeed(self, speed):
        '''Sets the speed of the game to a proportion of normal speed.
        That is, speed=2.0 is twice as fast a game as normal
        '''
        self._gameSpeed = speed
        self.tickPeriod = TICK_PERIOD * speed

    @GameOverMsg.handler
    def gameOver(self, msg):
        self._nextGameStateCheck = (
            self.timer.getElapsedTime() + RABBIT_CHASING_PERIOD)
        self.setWinningTeamById(msg.teamId)
        self.timer.gameFinishedNow()
        self.setGameState(gameStates.Ended)

        self.onGameOver(self.getTeam(msg.teamId), msg.timeOver)

        for player in self.players:
            player.readyToStart = False
            player.readyForTournament = False

        self._gameInProgress = False

    def gameIsInProgress(self):
        return self._gameInProgress

    @GameStartMsg.handler
    def gameStart(self, msg):
        self.setGameState(self.state.getInProgressState())
        self.timer.reset()
        self._gameInProgress = True
        for team in self.teams:
            team.resetScore()

    @AddPlayerMsg.handler
    def handle_AddPlayerMsg(self, msg):
        team = self.teamWithId[msg.teamId]
        zone = self.zoneWithId[msg.zoneId]

        # Create the player.
        nick = msg.nick.decode()
        player = Player(self, nick, team, msg.playerId, msg.dead, msg.bot)
        player.teleportToZoneCentre(zone)
        player.resyncBegun()

        self.addPlayer(player)

    def addPlayer(self, player):
        # Add this player to this universe.
        self.players.add(player)
        self.playerWithId[player.id] = player
        self.onPlayerAdded(player)

    @PlayerHasElephantMsg.handler
    def gotElephantMsg(self, msg):
        player = self.getPlayer(msg.playerId)
        self.playerWithElephant = player

    @PlayerHasTrosballMsg.handler
    def handle_PlayerHasTrosballMsg(self, msg):
        self.setTrosballPlayer(self.playerWithId[msg.playerId])

    def setTrosballPlayer(self, player):
        self.trosballPlayer = player
        self.trosballPlayer.deleteUpgrade()
        self.playerGotTrosballTime = self.timer.getElapsedTime()
        self.trosball = None
        self.trosballPlayer.onGotTrosball()

    @TrosballPositionMsg.handler
    def handle_TrosballPositionMsg(self, msg):
        self.setTrosballPosition((msg.xpos, msg.ypos), (msg.xvel, msg.yvel))

    def setTrosballPosition(self, pos, vel):
        self.trosballPlayer = None
        if self.trosball is None:
            self.trosball = Trosball(self)
        self.trosball.teleport(pos, vel[0], vel[1])

    @TrosballTagMsg.handler
    def handle_TrosballTagMsg(self, msg):
        self.setGameState(gameStates.BetweenTrosballPoints)
        if msg.teamId == '\x00':
            return
        self.trosballScore[msg.teamId] += 1
        self.trosballPlayer = None
        winningTeam = self.teamWithId[msg.teamId]
        netPosition = self.getTrosballTargetZoneDefn(winningTeam).pos
        if self.trosball is None:
            self.trosball = Trosball(self)
        self.trosball.setIsInNet(netPosition)

    def teamsWithHighestTrosballScore(self):
        maxScore = max(self.trosballScore.itervalues())
        return [
            self.teamWithId[teamId]
            for teamId, score in self.trosballScore.iteritems()
            if score == maxScore]

    def delPlayer(self, player):
        playerId = player.id
        player.removeFromGame()
        self.players.remove(player)
        del self.playerWithId[player.id]
        if player == self.playerWithElephant:
            self.returnElephantToOwner()

        # In case anyone else keeps a reference to it
        player.id = -1
        self.onPlayerRemoved(player, playerId)

    def advanceEverything(self):
        '''Advance the state of the game by deltaT seconds'''

        for shot in list(self.shots):
            if shot.expired:
                del self.shotWithId[shot.id]
                self.onShotRemoved(shot.id)

        # Update the player and shot positions.
        advancables = (
            self.shotWithId.values() + list(self.players) + list(self.grenades)
            + self.collectableStars.values())
        if self.trosball is not None:
            advancables.append(self.trosball)
        for unit in advancables:
            unit.reset()
            unit.advance()

        self.updateZoneInhabitants(advancables)

        if self.state.isTrosball():
            self.updateTrosballZones()

            if self.trosballPlayer is not None:
                if (self.timer.getElapsedTime() - self.playerGotTrosballTime >=
                        self.physics.trosballExplodeTime):
                    self.explodeTrosball()

    def getCollectableUnits(self):
        for star in self.collectableStars.values():
            yield star
        if self.trosball:
            yield self.trosball
        for unit in list(self.deadStars):
            yield unit

    def updateZoneInhabitants(self, advancables):
        for zone in self.map.zones:
            zone.clearPlayers()
        for unit in advancables:
            if isinstance(unit, Player):
                zone = unit.getZone()
                if zone:
                    zone.addPlayer(unit)

    def explodeTrosball(self):
        assert self.trosballPlayer is not None
        player = self.trosballPlayer
        player.killOutright(deathType=TROSBALL_DEATH)
        self.onTrosballExplosion(player)

    def updateTrosballZones(self):
        if self.state.isTrosball():
            try:
                trosballPosition = self.getTrosballPosition()
            except ValueError:
                pass
            else:
                for zone in self.zones:
                    # Don't update target zones:
                    if not any([
                            self.getTrosballTargetZoneDefn(team) ==
                            zone.defn for team in self.teams]):
                        zone.updateByTrosballPosition(trosballPosition)

    def bomberExploded(self, player):
        player.killOutright(deathType=BOMBER_DEATH, resync=False)
        self.onBomberExplosion(player)

    def canShoot(self):
        return self.physics.shooting and self.state.canShoot()

    def canRename(self):
        return self.state.canRename()

    def setWinningTeamById(self, teamId):
        self._winningTeam = self.teamWithId.get(teamId, None)

    def getWinningTeam(self):
        return self._winningTeam

    @CreateCollectableStarMsg.handler
    def createCollectableStar(self, msg):
        self.addCollectableStar(
            CollectableStar(self, msg.starId, (msg.xPos, msg.yPos)))

    @RemoveCollectableStarMsg.handler
    def gotRemoveCollectableStarMsg(self, msg):
        # On the server, the star must be remembered in case a player is about
        # to collect it but we don't know yet.
        if not self.isServer:
            del self.collectableStars[msg.starId]
            self.onCollectableStarRemoved(msg.starId)

    def teamWithAllZones(self):
        # Now check for an all zones win.
        team2Wins = self.teams[0].isLoser()
        team1Wins = self.teams[1].isLoser()
        if team1Wins and team2Wins:
            # The extraordinarily unlikely situation that all
            # zones have been neutralised in the same tick
            return 'Draw'
        elif team1Wins:
            return self.teams[0]
        elif team2Wins:
            return self.teams[1]
        else:
            return None

    def teamWithMoreZones(self):
        if self.teams[0].numZonesOwned > self.teams[1].numZonesOwned:
            return self.teams[0]
        elif self.teams[1].numZonesOwned > self.teams[0].numZonesOwned:
            return self.teams[1]
        else:
            return None

    def getTeamPlayerCounts(self):
        '''
        Returns a mapping from team id to number of players currently on that
        team.
        '''
        playerCounts = {}
        for player in self.players:
            playerCounts[player.teamId] = playerCounts.get(
                player.teamId, 0) + 1
        return playerCounts

    def getTeamToJoin(self, preferredTeamId):
        return self.state.getTeamToJoin(preferredTeamId, self)

    def getTeamName(self, id):
        if id == '\x00':
            return self.rogueTeamName
        return self.getTeam(id).teamName

    @FireShoxwaveMsg.handler
    def shoxwaveExplosion(self, msg):
        radius = 128
        # Get the player who fired this shoxwave
        shoxPlayer = self.getPlayer(msg.playerId)
        if not shoxPlayer:
            return
        shoxPlayer.weaponDischarged()

        # Loop through all the players in the game
        for player in self.players:
            if not (player.isFriendsWith(shoxPlayer) or distance(player.pos,
                    shoxPlayer.pos) > radius or player.dead or
                    player.isInvulnerable() or
                    player.phaseshift or player.turret):
                player.zombieHit(shoxPlayer, None, SHOXWAVE_DEATH)

        for shot in self.shotWithId.values():
            if (not shot.originatingPlayer.isFriendsWith(shoxPlayer) and
                    distance(shot.pos, shoxPlayer.pos) <= radius):
                shot.expired = True

    @ShotFiredMsg.handler
    def shotFired(self, msg):
        '''A player has fired a shot.'''
        try:
            player = self.playerWithId[msg.playerId]
        except KeyError:
            return

        shot = player.createShot(msg.shotId)
        self.shotWithId[msg.shotId] = shot
        player.weaponDischarged()
        player.onShotFired(shot)

    def _killTurret(self, tagger, zone):
        if zone.turretedPlayer is not None:
            zone.turretedPlayer.killOutright(
                deathType=TURRET_DEATH, killer=tagger)

    def trosballLost(self, formerOwner):
        self.setTrosballPosition(
            formerOwner.pos, formerOwner.getCurrentVelocity())

    @TaggingZoneMsg.handler
    def zoneTagged(self, msg):
        if msg.playerId == '\x00':
            player = None
        else:
            player = self.playerWithId[msg.playerId]
        zone = self.map.zoneWithId[msg.zoneId]
        previousOwner = zone.owner
        zone.tag(player)
        self._killTurret(player, zone)
        self.onZoneTagged(zone, player, previousOwner)
        if player:
            player.onTaggedZone(zone, previousOwner)

    @ZoneStateMsg.handler
    def zoneOwned(self, msg):
        zone = self.map.zoneWithId[msg.zoneId]
        team = self.teamWithId[msg.teamId]
        zone.setOwnership(team, msg.dark)

    @RespawnMsg.handler
    def respawn(self, msg):
        player = self.getPlayer(msg.playerId)
        zone = self.getZone(msg.zoneId)
        if player and zone:
            player.respawn(zone)

    @PlayerStarsSpentMsg.handler
    def starsSpent(self, msg):
        player = self.getPlayer(msg.playerId)
        if player:
            player.stars -= msg.count
            player.onStarsChanged()

    @UpgradeChangedMsg.handler
    def changeUpgrade(self, msg):
        for upgradeClass in allUpgrades:
            if upgradeClass.upgradeType == msg.upgradeType:
                if msg.statType == 'S':
                    upgradeClass.requiredStars = msg.newValue
                elif msg.statType == 'T':
                    upgradeClass.timeRemaining = msg.newValue
                elif msg.statType == 'E':
                    upgradeClass.enabled = bool(msg.newValue)

    def addGrenade(self, grenade):
        self.grenades.add(grenade)

    def removeGrenade(self, grenade):
        self.grenades.remove(grenade)

    def addCollectableStar(self, star):
        self.collectableStars[star.id] = star

    def setGameState(self, state):
        # Reset abilities and ui options
        self.abilities = Abilities()
        self.uiOptions = UIOptions()

        self.state = state
        self.state.applyTo(self)
        self.onGameStateChanged()
        self.state.onBecomingCurrentState(self)

    def getTeamStars(self, team):
        if team is None:
            return 0
        total = 0
        for p in self.players:
            if p.team == team:
                total += p.stars
        return total

    def getTrosballPosition(self):
        if self.trosball:
            return self.trosball.pos
        elif self.trosballPlayer:
            return self.trosballPlayer.pos
        elif self.state.isTrosball():
            # Game probably hasn't started yet, assume the middle.
            return (self.map.layout.centreX, self.map.layout.centreY)
        else:
            raise ValueError('no Trosball')

    def getTrosballTargetZoneDefn(self, team):
        return self.layout.getTrosballTargetZoneDefn(team)

    def setLayout(self, layout):
        self.zoneWithDef = {}
        for team in self.teams:
            team.numZonesOwned = 0
        self.layout = layout
        self.map = MapState(self, self.layout)
        self.zoneWithId = self.map.zoneWithId
        self.zones = self.map.zones
        self.zoneBlocks = self.map.zoneBlocks

    def setTestMode(self):
        self.PLAYER_RESET_STARS = 20
        for upgradetype in allUpgrades:
            upgradetype.requiredStars = 1

    def isTournamentTeam(self, teamName):
        if not self.authManager:
            return False
        return teamName in self.authManager.tournamentSettings.teamNames

    def getStartingGameStateFromString(self, gameType):
        if gameType == 'normal':
            return gameStates.Starting
        if gameType == 'trosball':
            return gameStates.StartingTrosball
        if gameType == 'solotrosball':
            return gameStates.StartingSoloTrosball
        if gameType == 'solo':
            return gameStates.StartingSolo
        raise ValueError('unknown game type %r', gameType)

    @TickMsg.handler
    def tickReceived(self, msg):
        self.timer.tick()
        self.advanceEverything()
        self.lastTickId = msg.tickId

        self.processDelayedCalls()

    def processDelayedCalls(self):
        now = self.timer.getElapsedTime()
        while self.delayedCalls:
            if self.delayedCalls[0][0] > now:
                break
            when, call = heapq.heappop(self.delayedCalls)
            if call.cancelled:
                continue
            try:
                call.fn(*call.args, **call.kwargs)
            except Exception:
                log.exception('Error in delayed call')

    @StartingSoonMsg.handler
    def startingSoon(self, msg):
        self.timer.beginGameStartCountdown(msg.delay)
        gameState = self.getStartingGameStateFromString(msg.gameType)
        self.setGameState(gameState)

    @ReturnToLobbyMsg.handler
    def returnToLobby(self, msg):
        self.lobbyTeamIds = {}
        self.setGameState(gameStates.Lobby)
        self._nextGameStateCheck = (
            self.timer.getElapsedTime() + GAME_STATE_CHECK_PERIOD)
        log.debug('Returning to lobby')

    def dumpEverything(self):
        '''Returns a dict representing the settings which must be sent to
        clients that connect to this server.'''

        result = {
            'loading': self.loading,
            'teams': [
                {
                    'id': team.id,
                    'name': team.teamName,
                    'trosballScore': self.trosballScore[team.id]
                        if self.state.isTrosball() else None,
                } for team in self.teams
            ],
            'worldMap': self.map.layout.toString(),
            'mode': self.gameMode,
            'state': self.state.id,
            'timer': self.timer.dumpState(),
            'speed': self._gameSpeed,
            'zones': [
                {
                    'id': zone.id,
                    'teamId': zone.owner.id if zone.owner else '\x00',
                    'dark': zone.dark,
                } for zone in self.zones
            ],
            'players': [player.dump() for player in self.players],
            'trosball': {
                'pos': self.trosball.pos if self.trosball else None,
                'vel': [self.trosball.xVel, self.trosball.yVel]
                    if self.trosball else None,
                'playerId': self.trosballPlayer.id
                    if self.trosballPlayer else None,
                'catchTime': self.playerGotTrosballTime
                    if self.trosballPlayer else None,
            } if self.trosball or self.trosballPlayer else None,
            'upgrades': [
                {
                    'type': upgrade.upgradeType,
                    'cost': upgrade.requiredStars,
                    'time': upgrade.timeRemaining,
                    'enabled': upgrade.enabled,
                } for upgrade in allUpgrades
            ],
            'elephant': self.playerWithElephant.id
                if self.playerWithElephant else None,
            'shots': [
                {
                    'id': shot.id,
                    'team': shot.team.id if shot.team else '\x00',
                    'pos': shot.pos,
                    'vel': shot.vel,
                    'shooter': shot.originatingPlayer.id,
                    'time': shot.timeLeft,
                    'kind': shot.kind,
                } for shot in self.shotWithId.itervalues() if not shot.expired
            ],
            'stars': [
                {
                    'id': star.id,
                    'created': star.creationTime,
                    'pos': star.pos,
                    'xVel': star.xVel,
                    'yVel': star.yVel,
                } for star in self.collectableStars.itervalues()
            ],
            'grenades': [
                {
                    'player': grenade.player.id if grenade.player else '\x00',
                    'pos': grenade.pos,
                    'xVel': grenade.xVel,
                    'yVel': grenade.yVel,
                    'timeLeft': grenade.timeLeft,
                } for grenade in self.grenades
            ],
            'physics': self.physics.dumpState(),
            'lastTickId': self.lastTickId,
        }

        return result

    def restoreEverything(self, data):
        self.loading = data['loading']
        if 'lastTickId' in data:
            self.lastTickId = data['lastTickId']

        for teamData in data['teams']:
            teamId = teamData['id']
            self.teamWithId[teamId].teamName = teamData['name']
            if teamData['trosballScore'] is not None:
                self.trosballScore[teamId] = teamData['trosballScore']

        gameState = gameStates.gameStatesById[data['state']]
        self.setGameState(gameState)
        self.timer.restoreState(data['timer'])
        self.physics.restoreState(data['physics'])

        mapSpec = data['worldMap']
        keys = MapLayout.unknownBlockKeys(self.layoutDatabase, mapSpec)
        if keys:
            # We don't know all of the map blocks, so there is no point
            # proceeding.
            self.isIncomplete = True
            self.onReset()
            raise UnknownMapLayouts(keys)

        layout = MapLayout.fromString(self.layoutDatabase, mapSpec)
        self.setLayout(layout)

        self.setGameMode(data['mode'])
        self.setGameSpeed(data['speed'])

        for zoneData in data['zones']:
            self.getZone(zoneData['id']).setOwnership(
                self.teamWithId[zoneData['teamId']], zoneData['dark'])

        for playerData in data['players']:
            playerId = playerData['id']
            team = self.teamWithId[playerData['teamId']]
            nick = playerData['nick']

            if playerId in self.playerWithId:
                player = self.playerWithId[playerId]
            else:
                player = Player(self, nick, team, playerId)
                self.addPlayer(player)

            player.restore(playerData)

        trosball = data['trosball']
        if trosball:
            if trosball['playerId']:
                self.setTrosballPlayer(
                    self.getPlayer(trosball['playerId']))
                self.playerGotTrosballTime = trosball['catchTime']
            elif trosball['pos']:
                self.setTrosballPosition(trosball['pos'], trosball['vel'])

        for upgradeData in data['upgrades']:
            for upgradeClass in allUpgrades:
                if upgradeClass.upgradeType == upgradeData['type']:
                    upgradeClass.requiredStars = upgradeData['cost']
                    upgradeClass.timeRemaining = upgradeData['time']
                    upgradeClass.enabled = upgradeData['enabled']

        self.playerWithElephant = self.getPlayer(data['elephant'])

        self.shotWithId = {}
        for shotData in data['shots']:
            shot = Shot(
                self, shotData['id'], self.getTeam(shotData['team']),
                self.getPlayer(shotData['shooter']), tuple(shotData['pos']),
                tuple(shotData['vel']), shotData['kind'], shotData['time'])
            self.shotWithId[shot.id] = shot

        self.deadStars = set()
        self.collectableStars = {}
        for starData in data['stars']:
            star = CollectableStar(
                self, starData['id'], tuple(starData['pos']))
            star.creationTime = starData['created']
            star.xVel = starData['xVel']
            star.yVel = starData['yVel']
            self.collectableStars[star.id] = star

        self.grenades = set()
        for grenadeData in data['grenades']:
            grenade = GrenadeShot(
                self, self.getPlayer(grenadeData['player']),
                grenadeData['timeLeft'])
            grenade.pos = grenadeData['pos']
            grenade.xVel = grenadeData['xVel']
            grenade.yVel = grenadeData['yVel']
            self.addGrenade(grenade)

        self.isIncomplete = False
        self.onReset()

    def canVote(self):
        return self.state == gameStates.Lobby and not self._onceOnly

    @ChangeTimeLimitMsg.handler
    def changeTimeLimit(self, msg):
        self.timer.setMatchDuration(msg.timeLimit)

    def resetTrosballToCentreOfMap(self):
        pass

    def elephantKill(self, killer):
        pass

    def playerHasNoticedDeath(self, player, killer):
        pass

    def stop(self):
        pass

    def callLater(self, _delay, _fn, *args, **kwargs):
        '''
        Schedules the given function to be called at the given game time.
        '''
        now = self.timer.getElapsedTime()
        heapq.heappush(
            self.delayedCalls, (now + _delay, DelayedCall(_fn, args, kwargs)))


class ServerUniverse(Universe):
    '''
    A universe that contains a few extra functions that are only needed
    server-side.
    '''

    isServer = True

    def __init__(self, game, *args, **kwargs):
        super(ServerUniverse, self).__init__(*args, **kwargs)
        self.game = game
        self.idManager = IdManager(self)
        self.starSpawner = StarSpawner(self)
        self._lastTickId = 0
        self.nextGameType = 'normal'
        self._loop = None
        self._startClock()

        reactor.callLater(0, self.loadPathFindingData)

        if __debug__ and globaldebug.enabled:
            globaldebug.serverUniverse = self

    @defer.inlineCallbacks
    def loadPathFindingData(self):
        from trosnoth.ais.pathfinding import RunTimePathFinder

        if self.map.layout.pathFinder:
            # Map has not changed since last load of data
            return

        self.sendServerCommand(WorldLoadingMsg(True))

        self.map.layout.pathFinder = RunTimePathFinder(self.map.layout)
        yield self.map.layout.pathFinder.loadData()
        self.sendServerCommand(WorldLoadingMsg(False))

    def needsPathFindingData(self, modeString):
        state = self.getStartingGameStateFromString(modeString)
        return state.aisEnabled()

    def sendServerCommand(self, msg):
        self.game.sendServerCommand(msg)

    def advanceEverything(self):
        super(ServerUniverse, self).advanceEverything()

        self.checkShotCollisions()
        self.updateZoneOwnership()
        self.starSpawner.tick()
        self.updateCollectableStars()
        for unit in list(self.deadStars):
            unit.beat()
            if not unit.history:
                self.deadStars.remove(unit)
                self.onCollectableStarRemoved(unit.id)
        for player in list(self.players):
            if player.resyncing and self.lastTickId > player.resyncExpiry:
                log.warning('%s took too long to resync', player)
                if player.agent:
                    player.agent.messageToAgent(ChatFromServerMsg(
                        error=True, text='You have been removed from the game '
                        'because your connection is too slow!'))
                self.sendServerCommand(RemovePlayerMsg(player.id))

    def checkShotCollisions(self, resolution=200):
        '''
        Performs collision checks of all shots with all nearby players.
        '''
        buckets = defaultdict(list)
        for player in self.players:
            if player.dead:
                continue
            x, y = player.pos
            xBucket, yBucket = x // resolution, y // resolution
            for dx in [-1, 0, 1]:
                for dy in [-1, 0, 1]:
                    buckets[xBucket + dx, yBucket + dy].append(player)

        for shot in list(self.shots):
            x, y = shot.pos
            xBucket, yBucket = x // resolution, y // resolution
            hitPlayers = [
                p for p in buckets[xBucket, yBucket]
                if shot.checkCollision(p)]
            if hitPlayers:
                # If multiple players are hit in the same tick, the server
                # randomly selects one to die.
                self.sendServerCommand(
                    ShotHitPlayerMsg(random.choice(hitPlayers).id, shot.id))

    def updateZoneOwnership(self):
        try:
            if not self.state.isTrosball():
                if not self.abilities.zoneCaps:
                    return
                tagger = None
                for zone in self.zones:
                    result = zone.playerWhoTaggedThisZone()
                    if result is not None:
                        tagger, team = result
                        if tagger is None:
                            assert team is None
                            self.sendServerCommand(
                                TaggingZoneMsg(zone.id, '\x00', '\x00'))
                        else:
                            self.sendServerCommand(TaggingZoneMsg(
                                zone.id, tagger.id, tagger.team.id))

                self.neutraliseDisconnectedZones(tagger)
        except:
            log.error('Error in zone ownership check', exc_info=True)

    def neutraliseDisconnectedZones(self, tagger):
        seenZones = set()
        teamSectors = dict((team, []) for team in self.teams)
        for zone in self.zones:
            if zone in seenZones:
                continue
            if zone.isNeutral():
                seenZones.add(zone)
                continue
            sector = zone.getContiguousZones()
            teamSectors[zone.owner].append(sector)
            seenZones.update(sector)

        for team, sectors in teamSectors.iteritems():
            if len(sectors) > 1:
                self.removeSmallestSector(team, sectors)
                zonesToNeutralise = list(itertools.chain(*sectors))
                self.neutraliseZones(zonesToNeutralise)
                self.onSectorNeutralised(tagger, len(zonesToNeutralise))
                if tagger:
                    # If two teams tag the same zone at the same time, tagger
                    # can be None.
                    tagger.onNeutralisedSector(len(zonesToNeutralise))

    def neutraliseZones(self, zones):
        for zone in zones:
            self.sendServerCommand(TaggingZoneMsg(zone.id, '\x00', '\x00'))

    def removeSmallestSector(self, team, sectors):
        '''
        Accepts team and list of sectors, selects which one that team should
        keep, returns the rest.
        '''
        key = functools.partial(self.getSectorGoodnessKey, team)
        goodSector = max(sectors, key=key)
        sectors.remove(goodSector)

    def getSectorGoodnessKey(self, team, sector):
        '''
        Returns a key by which sectors can be sorted with sectors which should
        be kept sorted as maximum.
        '''
        livePlayerCount = 0
        deadPlayerCount = 0
        darkZoneCount = 0
        for zone in sector:
            for player in zone.players:
                if player.team == team:
                    if player.dead:
                        deadPlayerCount += 1
                    else:
                        livePlayerCount += 1
            if zone.isDark():
                darkZoneCount += 1

        return (len(sector), livePlayerCount, darkZoneCount, deadPlayerCount)

    def updateCollectableStars(self):
        try:
            for star in self.collectableStars.values():
                if (star.creationTime < self.timer.getElapsedTime() -
                        COLLECTABLE_STAR_LIFETIME):
                    star.removeDueToTime()
        except:
            log.error('Error updating collectable stars', exc_info=True)

    def sendTrosballCollect(self, player):
        msg = PlayerHasTrosballMsg(player.id)
        self.sendServerCommand(msg)

    def sendTrosballPosition(self, x, y, xVel, yVel):
        msg = TrosballPositionMsg(x, y, xVel, yVel)
        self.sendServerCommand(msg)

    def resetTrosballToCentreOfMap(self):
        layout = self.map.layout
        x, y = (layout.centreX, layout.centreY)
        self.sendTrosballPosition(x, y, 0, 0)

    def throwTrosball(self):
        player = self.trosballPlayer
        xVel, yVel = player.getCurrentVelocity()
        angle = player.angleFacing
        xVel += self.physics.trosballThrowVel * sin(angle)
        yVel += -self.physics.trosballThrowVel * cos(angle)
        self.trosballCooldownPlayer = player

        WeakCallLater(0.2, self, 'clearTrosballCooldown', player)
        self.sendTrosballPosition(player.pos[0], player.pos[1], xVel, yVel)

    def clearTrosballCooldown(self, player):
        if self.trosballCooldownPlayer == player:
            self.trosballCooldownPlayer = None

    def returnElephantToOwner(self):
        for p in self.players:
            if p.isElephantOwner():
                self.sendServerCommand(PlayerHasElephantMsg(p.id))
                return
        self.sendServerCommand(PlayerHasElephantMsg('\x00'))

    def elephantKill(self, killer):
        if not killer.dead:
            self.sendServerCommand(PlayerHasElephantMsg(killer.id))
        else:
            self.returnElephantToOwner()

    def playerHasNoticedDeath(self, player, killer):
        if self.state.becomeNeutralWhenDie():
            if player.team is not None:
                self.sendServerCommand(SetPlayerTeamMsg(player.id, '\x00'))

        if killer and not killer.dead:
            self.sendServerCommand(AwardPlayerStarMsg(killer.id))

    @GameOverMsg.handler
    def gameOver(self, msg):
        super(ServerUniverse, self).gameOver(msg)

        for player in self.players:
            if player.dead and player.team is not None:
                self.sendServerCommand(SetPlayerTeamMsg(player.id, '\x00'))

    def _startGame(self, duration, size, delay=10):
        if self.loadedMap:
            layout = self.loadedMap
            self.loadedMap = None
        else:
            layout = self.layoutDatabase.generateRandomMapLayout(
                size[0], size[1])

        if duration is not None:
            self.timer.setMatchDuration(duration)
        self.setLayout(layout)

        self.beginGameStartCountdown(delay)

    def beginGameStartCountdown(self, delay=10):
        for player in self.players:
            zone = self.selectZoneForTeam(player.teamId)
            player.teleportToZoneCentre(zone)
            player.health = 0
            player.zombieHits = 0
            player.upgrade = None
            player._setStarsForDeath()
            player.respawnGauge = 0.0
            player.resyncBegun()

        self.trosballPlayer = None

        if self.state.isTrosball():
            self.trosball = Trosball(self)

        self.syncEverything()
        if self.needsPathFindingData(self.nextGameType):
            # By not yielding, we start the loading process and it continues on
            # its own
            self.loadPathFindingData()

        self.game.sendServerCommand(StartingSoonMsg(delay, self.nextGameType))

        self.callLater(delay, self.game.sendServerCommand, GameStartMsg())

    def syncEverything(self):
        self.game.sendServerCommand(WorldResetMsg(repr(self.dumpEverything())))

    def tick(self):
        if self.loading or len(self.players) == 0:
            return
        tickId = self._lastTickId = (self._lastTickId + 1) % TICK_LIMIT
        self.game.sendServerCommand(TickMsg(tickId))
        self.onServerTickComplete()

    def _startClock(self):
        if self._loop is not None:
            self._loop.stop()
        self._loop = WeakLoopingCall(self, 'tick')
        if __debug__ and globaldebug.enabled:
            self._loop.start(TICK_PERIOD * globaldebug.slowMotionFactor, False)
        else:
            self._loop.start(TICK_PERIOD, False)

    def _stopClock(self):
        if self._loop is not None:
            self._loop.stop()
            self._loop = None

    def stop(self):
        super(ServerUniverse, self).stop()
        self._stopClock()
        self.starSpawner.stop()

    @StartingSoonMsg.handler
    def startingSoon(self, msg):
        super(ServerUniverse, self).startingSoon(msg)

        # Kick all AI players now, the game is about to begin.
        for p in list(self.players):
            if p.bot:
                self.game.sendServerCommand(RemovePlayerMsg(p.id))

    def checkForResult(self):
        msg = self.state.getFinishedMessage(self)
        if msg is not None:
            self.game.sendServerCommand(msg)

    @TickMsg.handler
    def tickReceived(self, msg):
        super(ServerUniverse, self).tickReceived(msg)

        self.checkForResult()
        self.changeGameStateIfNeeded()
        if self.state == gameStates.Ended:
            self.checkForRabbitsAllDead()

    def changeGameStateIfNeeded(self):
        if ((self.state not in (gameStates.Lobby, gameStates.Ended)) or
                self._onceOnly):
            return
        if self.timer.getElapsedTime() < self._nextGameStateCheck:
            return
        self._nextGameStateCheck = (
            self.timer.getElapsedTime() + GAME_STATE_CHECK_PERIOD)

        if self.state == gameStates.Ended:     # At least one rabbit survived!
            log.debug('Time to return to lobby')
            self.game.sendServerCommand(ReturnToLobbyMsg())
            return
        assert self.state == gameStates.Lobby
        self.startNewGameIfReady()

    def startNewGameIfReady(self):
        if self.layoutDatabase is None:
            return
        self._gameStarter.startNewGameIfReady()

    def setNextGameType(self, gameType):
        '''
        @param gameType: string representing next game type.
        '''
        self.getStartingGameStateFromString(gameType)
        self.nextGameType = gameType

    def getNextGameType(self):
        return self.nextGameType

    def checkForRabbitsAllDead(self):
        log.debug('checking for rabbits all dead')
        if all(p.team is None for p in self.players):
            self.game.sendServerCommand(ReturnToLobbyMsg())

    @TrosballTagMsg.handler
    def handle_TrosballTagMsg(self, msg):
        super(ServerUniverse, self).handle_TrosballTagMsg(msg)
        WeakCallLater(
            4, self, '_startGame', None, (
                self.layout.halfMapWidth,
                self.layout.mapHeight),
            4)

    def setGameState(self, state):
        self._gameStarter = state.getGameStarter(self, self._startGame)
        super(ServerUniverse, self).setGameState(state)

    @WorldResetMsg.handler
    def gotWorldReset(self, msg):
        super(ServerUniverse, self).gotWorldReset(msg)
        self.starSpawner.reset()
        self.onReset()

    @GameStartMsg.handler
    def gameStart(self, msg):
        super(ServerUniverse, self).gameStart(msg)
        self.starSpawner.reset()



'''upgrades.py - defines the behaviour of upgrades.'''

import logging

import pygame

from trosnoth.model.shot import GrenadeShot

log = logging.getLogger('upgrades')

upgradeOfType = {}
allUpgrades = set()


def registerUpgrade(upgradeClass):
    '''
    Marks the given upgrade class to be used in the game.
    '''
    specialUpgrade(upgradeClass)
    allUpgrades.add(upgradeClass)

    return upgradeClass


def specialUpgrade(upgradeClass):
    '''
    Marks the given upgrade class as a special upgrade that can be used only
    via the console.
    '''
    if upgradeClass.upgradeType in upgradeOfType:
        raise KeyError('2 upgrades with %r' % (upgradeClass.upgradeType,))
    upgradeOfType[upgradeClass.upgradeType] = upgradeClass

    return upgradeClass


class Upgrade(object):
    '''Represents an upgrade that can be bought.'''
    goneWhenDie = True
    defaultKey = None
    iconPath = None
    iconColourKey = None
    enabled = True
    applyLocally = False

    # If this flag is set, instead of waiting for a verdict from the server,
    # the client will guess whether a purchase of this upgrade will be allowed,
    # and go on that assumption until it finds out otherwise. This exists so
    # that grenades can feel responsive even if there's high latency.
    doNotWaitForServer = False

    # Upgrades have an upgradeType: this must be a unique, single-character
    # value.
    def __init__(self, player):
        self.universe = player.world
        self.player = player
        self.verified = False

    def __str__(self):
        return self.name

    def use(self):
        '''Initiate the upgrade (server-side)'''
        pass

    def localUse(self, localState):
        '''Initiate the upgrade (client-side)'''
        if self.applyLocally:
            self.use()

    def serverVerified(self, localState):
        '''
        Called client-side when the server has confirmed that this upgrade is
        allowed. This is most useful as a hook for upgrade classes which set
        the doNotWaitForServer flag.
        '''
        self.verified = True

    def deniedByServer(self, localState):
        '''
        Called client-side when a doNotWaitForServer upgrade needs to be
        canceled because the server did not approve it.
        '''
        self.delete()

    def delete(self):
        '''
        Performs any necessary tasks to remove this upgrade from the game.
        '''
        self.player.upgrade = None

    def timeIsUp(self):
        '''
        Called by the universe when the upgrade's time has run out.
        '''
        pass


@registerUpgrade
class Shield(Upgrade):
    '''
    shield: protects player from one shot
    '''
    upgradeType = 's'
    requiredStars = 4
    timeRemaining = 30
    name = 'Shield'
    action = 'shield'
    order = 20
    defaultKey = pygame.K_2
    iconPath = 'upgrade-shield.png'

    def __init__(self, player):
        super(Shield, self).__init__(player)
        self.maxProtections = self.universe.physics.playerRespawnHealth
        self.protections = self.maxProtections


@specialUpgrade
class PhaseShift(Upgrade):
    '''
    phase shift: affected player cannot be shot, but cannot shoot.
    '''
    upgradeType = 'h'
    requiredStars = 6
    timeRemaining = 25
    name = 'Phase Shift'
    action = 'phase shift'
    order = 40
    defaultKey = pygame.K_4
    iconPath = 'upgrade-phaseshift.png'


@specialUpgrade
class Turret(Upgrade):
    '''
    turret: turns a player into a turret; a more powerful player, although one
    who is unable to move.
    '''
    upgradeType = 't'
    requiredStars = 8
    timeRemaining = 50
    name = 'Turret'
    action = 'turret'
    order = 10
    defaultKey = pygame.K_1
    iconPath = 'upgrade-turret.png'

    def use(self):
        '''Initiate the upgrade'''
        self.player.detachFromEverything()
        self.zone = zone = self.player.getZone()
        if not zone:
            return
        if zone.turretedPlayer is not None:
            zone.turretPlayer.upgrade.delete()
        zone.turretedPlayer = self.player

        # Arrest vertical movement so that upon losing the upgrade, the
        # player doesn't re-jump
        self.player.yVel = 0

    def delete(self):
        '''Performs any necessary tasks to remove this upgrade from the game'''
        self.player.turretOverHeated = False
        self.player.turretHeat = 0.0
        if self.zone:
            self.zone.turretedPlayer = None
        super(Turret, self).delete()


@registerUpgrade
class MinimapDisruption(Upgrade):
    upgradeType = 'm'
    requiredStars = 15
    timeRemaining = 40
    name = 'Minimap Disruption'
    action = 'minimap disruption'
    order = 30
    defaultKey = pygame.K_3
    iconPath = 'upgrade-minimap.png'

    def use(self):
        if self.player.team is not None:
            self.player.team.usingMinimapDisruption = True

    def delete(self):
        if self.player.team is not None:
            self.player.team.usingMinimapDisruption = False
            super(MinimapDisruption, self).delete()


@registerUpgrade
class Grenade(Upgrade):
    upgradeType = 'g'
    requiredStars = 7
    timeRemaining = 2.5
    goneWhenDie = False
    numShots = 40
    name = 'Grenade'
    action = 'grenade'
    order = 50
    defaultKey = pygame.K_5
    iconPath = 'upgrade-grenade.png'
    doNotWaitForServer = True

    def localUse(self, localState):
        localState.grenadeLaunched()

    def serverVerified(self, localState):
        super(Grenade, self).serverVerified(localState)
        localState.matchGrenade()

    def deniedByServer(self, localState):
        super(Grenade, self).deniedByServer(localState)
        localState.grenadeRemoved()

    def use(self):
        '''Initiate the upgrade.'''
        self.player.world.addGrenade(
            GrenadeShot(self.player.world, self.player, self.timeRemaining))


@registerUpgrade
class Ricochet(Upgrade):
    upgradeType = 'r'
    requiredStars = 3
    timeRemaining = 10
    name = 'Ricochet'
    action = 'ricochet'
    order = 60
    defaultKey = pygame.K_6
    iconPath = 'upgrade-ricochet.png'


@registerUpgrade
class Ninja (Upgrade):
    '''allows you to become invisible to all players on the opposing team'''
    upgradeType = 'n'
    requiredStars = 5
    timeRemaining = 25
    name = 'Ninja'
    action = 'phase shift'  # So old phase shift hotkeys trigger ninja.
    order = 40
    defaultKey = pygame.K_4
    iconPath = 'upgrade-ninja.png'
    iconColourKey = (254, 254, 253)


@registerUpgrade
class Shoxwave(Upgrade):
    '''
    shockwave: upgrade that will replace shots with a shockwave like that of
    the grenade vaporising all enemies and enemy shots in the radius of blast.
    '''
    upgradeType = 'w'
    requiredStars = 7
    timeRemaining = 45
    name = 'Shoxwave'
    action = 'shoxwave'
    order = 80
    defaultKey = pygame.K_7
    iconPath = 'upgrade-shoxwave.png'
    iconColourKey = (255, 255, 255)


@registerUpgrade
class MachineGun(Upgrade):
    upgradeType = 'x'
    requiredStars = 10
    timeRemaining = 30
    name = 'Machine Gun'
    action = 'turret'   # So that old turret hotkeys trigger machine gun.
    order = 10
    defaultKey = pygame.K_1
    iconPath = 'upgrade-machinegun.png'
    applyLocally = True

    def use(self):
        self.player.mgBulletsRemaining = 15
        self.player.mgFiring = False

    def delete(self):
        super(MachineGun, self).delete()
        self.player.mgBulletsRemaining = 0


@registerUpgrade
class Bomber(Upgrade):
    upgradeType = 'b'
    requiredStars = 1
    timeRemaining = 6
    name = 'Bomber'
    action = 'bomber'
    order = 90
    defaultKey = pygame.K_8
    iconPath = 'upgrade-bomber.png'

    def timeIsUp(self):
        super(Bomber, self).delete()
        self.universe.bomberExploded(self.player)


@specialUpgrade
class RespawnFreezer(Upgrade):
    '''
    Respawn freezer: upgrade that will render spawn points unusable.
    '''
    upgradeType = 'f'
    requiredStars = 8
    timeRemaining = 30
    name = 'Respawn Freezer'
    action = 'respawn freezer'
    order = 100
    defaultKey = pygame.K_9
    iconPath = 'upgrade-freezer.png'

    def use(self):
        '''Initiate the upgrade'''
        self.zone = self.player.getZone()
        self.zone.frozen = True

    def delete(self):
        '''Performs any necessary tasks to remove this upgrade from the game'''
        super(RespawnFreezer, self).delete()
        if self.zone:
            self.zone.frozen = False


@specialUpgrade
class Directatorship(Upgrade):
    '''
    Directatorship: shielded ninja ricochet machine gunner.
    '''
    upgradeType = 'd'
    requiredStars = 150
    timeRemaining = 60
    name = 'Directatorship'
    action = 'directatorship'
    order = 200
    maxProtections = protections = 5



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from math import pi
import random

from trosnoth.ai import AI
from trosnoth.utils.twist import WeakCallLater, WeakLoopingCall

class SimpleAI(AI):
    '''
    An example of a simple AI which runs around randomly and shoots at the
    nearest enemy.
    '''
    nick = 'SimpleBot'
    playable = True

    def start(self):
        self.stopped = False
        self._pauseTimer = None
        self._loop = WeakLoopingCall(self, 'tick')
        self._loop.start(0.5)

    def disable(self):
        self.stopped = True
        self._loop.stop()
        if self._pauseTimer is not None:
            self._pauseTimer.cancel()

    def tick(self):
        if self.stopped:
            return
        if self.player.dead:
            if self.player.inRespawnableZone():
                self.doAimAt(0, thrust=0)
                if self.player.respawnGauge == 0:
                    self.doRespawn()
            else:
                self.aimAtFriendlyZone()
        else:
            if self._pauseTimer is None:
                self.startPlayerMoving()
            if self.player.canShoot():
                self.fireAtNearestEnemy()

    def aimAtFriendlyZone(self):
        zones = [z for z in self.world.zones if z.owner == self.player.team]
        if len(zones) == 0:
            return

        def getZoneDistance(zone):
            x1, y1 = self.player.pos
            x2, y2 = zone.defn.pos
            return ((x2 - x1) ** 2 + (y2 - y1) ** 2) ** 0.5

        bestZone = min(zones, key=getZoneDistance)
        self.doAimAtPoint(bestZone.defn.pos)

    def fireAtNearestEnemy(self):
        enemies = [p for p in self.world.players if not
                (p.dead or p.invisible or self.player.isFriendsWith(p))]
        if len(enemies) == 0:
            return

        def getPlayerDistance(p):
            x1, y1 = self.player.pos
            x2, y2 = p.pos
            return ((x2 - x1) ** 2 + (y2 - y1) ** 2) ** 0.5

        nearestEnemy = min(enemies, key=getPlayerDistance)
        if getPlayerDistance(nearestEnemy) < 512:
            self.doAimAtPoint(nearestEnemy.pos)
            self.doShoot()

    def died(self, killer, deathType):
        self._pauseTimer = None
        self.doStop()
        self.aimAtFriendlyZone()

    def respawned(self):
        self.startPlayerMoving()

    def startPlayerMoving(self):
        self.pauseAndReconsider()

    def pauseAndReconsider(self):
        if self.stopped:
            return
        if self.player.dead:
            self._pauseTimer = None
            return

        # Pause again in between 0.5 and 2.5 seconds time.
        t = random.random() * 2 + 0.5
        self._pauseTimer = WeakCallLater(t, self, 'pauseAndReconsider')

        # Decide whether to jump or drop.
        if self.player.isAttachedToWall():
            verticalActions = ['none', 'jump', 'drop']
        elif self.player.isOnPlatform():
            verticalActions = ['none', 'jump', 'drop']
        elif self.player.isOnGround():
            verticalActions = ['none', 'jump']
        else:
            verticalActions = ['none']
        action = random.choice(verticalActions)
        if action == 'jump':
            self.doJump()
        elif action == 'drop':
            self.doDrop()

        # Decide on a direction.
        d = random.choice(['left', 'right', 'stop'])
        if d == 'left':
            self.doAimAt(-pi/2.)
            self.doMoveLeft()
        elif d == 'right':
            self.doAimAt(pi/2.)
            self.doMoveRight()
        else:
            self.doStop()

AIClass = SimpleAI



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from trosnoth.gui.app import MultiWindowApplication
from trosnoth.gui.framework import framework, elements, hotkey, prompt
from trosnoth.gui.framework.dialogbox import MoveableBox, DialogBox, YesNoBox, YesNoCancelBox, OkCancelBox, OkBox, DialogResult
from trosnoth.gui.common import Area, ScaledSize
from trosnoth.gui.fonts.font import ScaledFont
import pygame

class Interface(framework.CompoundElement):
    def __init__(self, app):
        super(Interface, self).__init__(app)

        # Draw the background
        bg = pygame.surface.Surface(app.screenManager.scaledSize)
        bg.fill((0,128,0))
        from random import randint
        for x in range(0,20):
            randPos = (randint(0,580), randint(0,430))
            pygame.draw.rect(bg, (255,0,0), pygame.rect.Rect(randPos, (20,20)), 10)
        box = FunBox(app, "Disgusting Colour Scheme?")
        box._setPos((0,0))

        box2 = FunBox(app, "Heck Yes!")
        box2.setColours(borderColour = (192,0,192))
        box3 = FunBox(app, "")
        box3._setPos((10,10))
        box4 = YesNoBox(app, (300,200), "A question: ", "Is Trosnoth cool?")
        box4.onClose.addListener(lambda:OkBox(app, (200,100), "Result",
                "You pressed %s" % DialogResult.text[box4.result]).show())
        self.elements = [elements.PictureElement(app, bg)]
        box.onClose.addListener(lambda:self.app.screenManager.setScreenProperties((1024,768), 0, "Testing"))
        box3.show()
        box.show()
        box2.show()
        box4.show()
        
        
##class FunBox(DialogBox):
class FunBox(MoveableBox):
    def __init__(self, app, textBoxText):
        
        super(FunBox, self).__init__(app, ScaledSize(300,200), "Dialog!!")
        h = prompt.InputBox(app, Area((10,10), (280,50)), textBoxText, font = app.screenManager.fonts.bigMenuFont)
        h.onClick.addListener(self.setFocus)
        font = ScaledFont("KLEPTOCR.TTF", 30)
        b = elements.TextButton(app, (150, 100), "Close", font, (255,0,0), (20,100,200))
        b.onClick.addListener(lambda sender: self.close())
        self.elements = [h, b]
        self.setColours(titleColour = (255,0,0), backgroundColour = (255,128,0))

size = (600,450)        
        
if __name__ == "__main__":        
    a = MultiWindowApplication(size, 0, "Testing", Interface)
    a.run()
